#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# BiblelatorSettingsEditor.py
#
# BOS (Bible Organisational System) manager program
#
# Copyright (C) 2016-2020 Robert Hunt
# Author: Robert Hunt <Freely.Given.org+Biblelator@gmail.com>
# License: See gpl-3.0.txt
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Program to allow viewing of various BOS (Bible Organisational System) subsystems
    such as versification systems, books names systems, etc.
"""
from gettext import gettext as _
import sys
import os
import logging
import subprocess
import multiprocessing

import tkinter as tk
from tkinter.ttk import Style, Frame, Button, Scrollbar, Label, Notebook
from tkinter.scrolledtext import ScrolledText

# BibleOrgSys imports
from BibleOrgSys import BibleOrgSysGlobals
from BibleOrgSys.BibleOrgSysGlobals import vPrint
from BibleOrgSys.Reference.BibleOrganisationalSystems import BibleOrganisationalSystem
from BibleOrgSys.Reference.BibleStylesheets import BibleStylesheet

# Biblelator imports
if __name__ == '__main__':
    aboveAboveFolderPath = os.path.dirname( os.path.dirname( os.path.dirname( os.path.abspath( __file__ ) ) ) )
    if aboveAboveFolderPath not in sys.path:
        sys.path.insert( 0, aboveAboveFolderPath )
from Biblelator.BiblelatorGlobals import DEFAULT, tkSTART, MAX_RECENT_FILES, errorBeep, \
        DATAFILES_FOLDERPATH, \
        DATA_FOLDER_NAME, LOGGING_SUBFOLDER_NAME, SETTINGS_SUBFOLDER_NAME, \
        DEFAULT_KEY_BINDING_DICT, MAX_PSEUDOVERSES, \
        parseWindowGeometry, assembleWindowGeometryFromList, centreWindow, \
        parseWindowSize
from Biblelator.Dialogs.BiblelatorSimpleDialogs import showError, showWarning, showInfo
from Biblelator.Dialogs.BiblelatorDialogs import SelectResourceBoxDialog, GetNewProjectNameDialog, \
                                CreateNewProjectFilesDialog, GetNewCollectionNameDialog, \
                                BookNameDialog, NumberButtonDialog
from Biblelator.Helpers.BiblelatorHelpers import mapReferencesVerseKey, createEmptyUSFMBooks
from Biblelator.Settings.Settings import ApplicationSettings, ProjectSettings
from Biblelator.Settings.BiblelatorSettingsFunctions import parseAndApplySettings, writeSettingsFile, \
        saveNewWindowSetup, deleteExistingWindowSetup, applyGivenWindowsSettings, viewSettings
from Biblelator.Windows.TextBoxes import BEntry, BCombobox
from Biblelator.Windows.ChildWindows import ChildWindows
from Biblelator.Windows.TextEditWindow import TextEditWindow


LAST_MODIFIED_DATE = '2020-04-26' # by RJH
SHORT_PROGRAM_NAME = "BiblelatorSettingsEditor"
PROGRAM_NAME = "Biblelator Settings Editor"
PROGRAM_VERSION = '0.46'
programNameVersion = f'{SHORT_PROGRAM_NAME} v{PROGRAM_VERSION}'

debuggingThisModule = False


MAIN_APP_NAME = 'Biblelator'
# Default window size settings (Note: X=width, Y=height)
INITIAL_MAIN_SIZE, INITIAL_MAIN_SIZE_DEBUG, MINIMUM_MAIN_SIZE, MAXIMUM_MAIN_SIZE = '607x376', '607x460', '550x375', '700x600'



class BiblelatorSettingsEditor( Frame ):
    """
    This is the main application window (well, actually a frame in the root toplevel window).

    Its main job is to keep track of self.currentVerseKey (and self.currentVerseKeyGroup)
        and use that to inform child windows of BCV movements.
    """
    global settings
    def __init__( self, rootWindow, homeFolderPath, loggingFolderPath, iconImage ):
        """
        Main app initialisation function.

        Creates the main menu and toolbar which includes the main BCV (book/chapter/verse) selector.
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "BiblelatorSettingsEditor.__init__( {}, {}, {}, … )".format( rootWindow, homeFolderPath, loggingFolderPath ) )
        self.rootWindow, self.homeFolderPath, self.loggingFolderPath, self.iconImage = rootWindow, homeFolderPath, loggingFolderPath, iconImage
        self.parentApp = self # Yes, that's me, myself!
        self.starting = True

        self.themeName = 'default'
        self.style = Style()
        self.interfaceLanguage = DEFAULT
        self.interfaceComplexity = DEFAULT
        self.touchMode = False # True makes larger buttons
        self.tabletMode = False
        self.showDebugMenu = False
        self.internetAccessEnabled = False

        self.lastFind = None
        #self.openDialog = None
        self.saveDialog = None
        self.optionsDict = {}

        self.fixedSettingsFlag = False # Can the user change the settings file that we're looking at?

        if BibleOrgSysGlobals.debugFlag: vPrint( 'Quiet', debuggingThisModule, "Button default font", Style().lookup('TButton', 'font') )
        if BibleOrgSysGlobals.debugFlag: vPrint( 'Quiet', debuggingThisModule, "Label default font", Style().lookup('TLabel', 'font') )

        self.stylesheet = BibleStylesheet().loadDefault()
        super().__init__( self.rootWindow )
        self.pack()

        self.rootWindow.protocol( 'WM_DELETE_WINDOW', self.doCloseMe ) # Catch when app is closed

        self.childWindows = ChildWindows( self )

        self.createStatusBar()

        ## Create our display text box
        #self.textBox = ScrolledText( self.rootWindow, bg='yellow' )#style='DebugText.TScrolledText' )
        #self.textBox.pack( side=tk.TOP, fill=tk.BOTH )
        ##self.debugTextBox.tag_configure( 'emp', background='yellow', font='helvetica 12 bold', relief='tk.RAISED' )
        #self.textBox.tag_configure( 'emp', font='helvetica 10 bold' )
        #self.textBox.insert( tk.END, 'Main Text Box:' )

        if BibleOrgSysGlobals.debugFlag: # Create a scrolling debug box
            self.lastDebugMessage = None
            #Style().configure('DebugText.TScrolledText', padding=2, background='orange')
            self.debugTextBox = ScrolledText( self.rootWindow, bg='orange' )#style='DebugText.TScrolledText' )
            self.debugTextBox.pack( side=tk.BOTTOM, fill=tk.BOTH )
            #self.debugTextBox.tag_configure( 'emp', background='yellow', font='helvetica 12 bold', relief='tk.RAISED' )
            self.debugTextBox.tag_configure( 'emp', font='helvetica 10 bold' )
            if debuggingThisModule: self.setDebugText( "Starting up…" )

        self.keyBindingDict = DEFAULT_KEY_BINDING_DICT
        self.myKeyboardBindingsList = []
        self.recentFiles = []

        # Read and apply the saved settings
        if BibleOrgSysGlobals.commandLineArguments.override is None:
            self.INIname = MAIN_APP_NAME # We use the Biblelator settings
            if BibleOrgSysGlobals.debugFlag and debuggingThisModule: vPrint( 'Quiet', debuggingThisModule, "Using default {!r} ini file".format( self.INIname ) )
        else:
            self.INIname = BibleOrgSysGlobals.commandLineArguments.override
            vPrint( 'Normal', debuggingThisModule, _("Using settings from user-specified {!r} ini file").format( self.INIname ) )
        #self.settings = ApplicationSettings( self.homeFolderPath, DATA_FOLDER_NAME, SETTINGS_SUBFOLDER_NAME, self.INIname )
        #self.settings.load()
        #if PROGRAM_NAME not in self.settings.data or 'windowSize' not in self.settings.data[PROGRAM_NAME] or 'windowPosition' not in self.settings.data[PROGRAM_NAME]:
        initialMainSize = INITIAL_MAIN_SIZE_DEBUG if BibleOrgSysGlobals.debugFlag else INITIAL_MAIN_SIZE
        centreWindow( self.rootWindow, *initialMainSize.split( 'x', 1 ) )

        if self.touchMode:
            vPrint( 'Normal', debuggingThisModule, _("Touch mode enabled!") )
            self.createTouchMenuBar()
            self.createTouchNavigationBar()
        else: # assume it's regular desktop mode
            self.createNormalMenuBar()
            self.createNormalNavigationBar()
        self.createToolBar()
        if BibleOrgSysGlobals.debugFlag: self.createDebugToolBar()
        self.createMainKeyboardBindings()
        self.createMainButtons()
        self.createNotebook()

        # See if there's any developer messages
        if self.internetAccessEnabled and self.checkForDeveloperMessagesEnabled:
            self.doCheckForDeveloperMessages()

        self.rootWindow.title( programNameVersion )
        self.minimumSize = MINIMUM_MAIN_SIZE
        self.rootWindow.minsize( *parseWindowSize( self.minimumSize ) )
        if BibleOrgSysGlobals.debugFlag: self.setDebugText( "__init__ finished." )
        self.starting = False
        self.setReadyStatus()
    # end of BiblelatorSettingsEditor.__init__


    def setGenericBibleOrganisationalSystem( self, BOSname ):
        """
        We usually use a fairly generic BibleOrganisationalSystem (BOS) to ensure
            that it contains all the books that we might ever want to navigate to.
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "setGenericBibleOrganisationalSystem( {} )".format( BOSname ) )

        # Set-up our Bible system and our callables
        self.genericBibleOrganisationalSystem = BibleOrganisationalSystem( self.genericBibleOrganisationalSystemName )
        self.genericBookList = self.genericBibleOrganisationalSystem.getBookList()
        #self.getNumBooks = self.genericBibleOrganisationalSystem.getNumBooks
        self.getNumChapters = self.genericBibleOrganisationalSystem.getNumChapters
        self.getNumVerses = lambda BBB,C: MAX_PSEUDOVERSES if C=='-1' or C==-1 \
                                        else self.genericBibleOrganisationalSystem.getNumVerses( BBB, C )
        self.isValidBCVRef = self.genericBibleOrganisationalSystem.isValidBCVRef
        self.getFirstBookCode = self.genericBibleOrganisationalSystem.getFirstBookCode
        self.getPreviousBookCode = self.genericBibleOrganisationalSystem.getPreviousBookCode
        self.getNextBookCode = self.genericBibleOrganisationalSystem.getNextBookCode
        self.getBBBFromText = self.genericBibleOrganisationalSystem.getBBBFromText
        self.getGenericBookName = self.genericBibleOrganisationalSystem.getBookName
        #self.getBookList = self.genericBibleOrganisationalSystem.getBookList

        # Make a bookNumber table with GEN as #1
        #vPrint( 'Quiet', debuggingThisModule, self.genericBookList )
        self.offsetGenesis = self.genericBookList.index( 'GEN' )
        #vPrint( 'Quiet', debuggingThisModule, 'offsetGenesis', self.offsetGenesis )
        self.bookNumberTable = {}
        for j,BBB in enumerate(self.genericBookList):
            k = j + 1 - self.offsetGenesis
            nBBB = BibleOrgSysGlobals.loadedBibleBooksCodes.getReferenceNumber( BBB )
            #vPrint( 'Quiet', debuggingThisModule, BBB, nBBB )
            self.bookNumberTable[k] = BBB
            self.bookNumberTable[BBB] = k
        #vPrint( 'Quiet', debuggingThisModule, self.bookNumberTable )
    # end of BiblelatorSettingsEditor.setGenericBibleOrganisationalSystem


    def createNormalMenuBar( self ):
        """
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "createNormalMenuBar()" )

        #self.win = Toplevel( self )
        self.menubar = tk.Menu( self.rootWindow )
        #self.rootWindow['menu'] = self.menubar
        self.rootWindow.configure( menu=self.menubar ) # alternative

        fileMenu = tk.Menu( self.menubar, tearoff=False )
        self.menubar.add_cascade( menu=fileMenu, label=_('File'), underline=0 )
        ##fileMenu.add_command( label=_('New…'), underline=0, command=self.notWrittenYet )
        #fileNewSubmenu = tk.Menu( fileMenu, tearoff=False )
        #fileMenu.add_cascade( label=_('New'), underline=0, menu=fileNewSubmenu )
        #fileNewSubmenu.add_command( label=_('Text file'), underline=0, command=self.notWrittenYet )
        #fileOpenSubmenu = tk.Menu( fileMenu, tearoff=False )
        #fileMenu.add_cascade( label=_('Open'), underline=0, menu=fileOpenSubmenu )
        #fileRecentOpenSubmenu = tk.Menu( fileOpenSubmenu, tearoff=False )
        #fileOpenSubmenu.add_cascade( label=_('Recent'), underline=0, menu=fileRecentOpenSubmenu )
        #for j, (filename, folder, windowType) in enumerate( self.recentFiles ):
            #fileRecentOpenSubmenu.add_command( label=filename, underline=0, command=self.notWrittenYet )
        #fileOpenSubmenu.add_separator()
        #fileOpenSubmenu.add_command( label=_('Text file…'), underline=0, command=self.notWrittenYet )
        #fileMenu.add_separator()
        #fileMenu.add_command( label=_('Save all…'), underline=0, command=self.notWrittenYet )
        #fileMenu.add_separator()
        #fileMenu.add_command( label=_('Save settings'), underline=0, command=self.notWrittenYet )
        #fileMenu.add_separator()
        fileMenu.add_command( label=_('Quit app'), underline=0, command=self.doCloseMe, accelerator=self.keyBindingDict[_('Quit')][0] ) # quit app

        # if 0:
        #     #editMenu = tk.Menu( self.menubar, tearoff=False )
        #     #self.menubar.add_cascade( menu=editMenu, label=_('Edit'), underline=0 )
        #     #editMenu.add_command( label=_('Find…'), underline=0, command=self.notWrittenYet )
        #     #editMenu.add_command( label=_('Replace…'), underline=0, command=self.notWrittenYet )

        #     booksCodesMenu = tk.Menu( self.menubar, tearoff=False )
        #     self.menubar.add_cascade( menu=booksCodesMenu, label=_('Codes'), underline=0 )
        #     booksCodesMenu.add_command( label=_('View…'), underline=-1, command=self.notWrittenYet )
        #     booksCodesMenu.add_separator()
        #     booksCodesMenu.add_command( label=_('Info…'), underline=0, command=self.doGotoInfo )

        #     ordersMenu = tk.Menu( self.menubar, tearoff=False )
        #     self.menubar.add_cascade( menu=ordersMenu, label=_('Orders'), underline=0 )
        #     ordersMenu.add_command( label=_('New…'), underline=0, command=self.notWrittenYet )
        #     ordersMenu.add_separator()
        #     ordersMenu.add_command( label=_('Info…'), underline=0, command=self.notWrittenYet )

        #     namesMenu = tk.Menu( self.menubar, tearoff=False )
        #     self.menubar.add_cascade( menu=namesMenu, label=_('Names'), underline=0 )
        #     submenuBibleResourceType = tk.Menu( namesMenu, tearoff=False )
        #     namesMenu.add_command( label=_('View…'), underline=5, command=self.notWrittenYet )
        #     namesMenu.add_separator()
        #     namesMenu.add_command( label=_('Info…'), underline=0, command=self.notWrittenYet )

        #     versificationsMenu = tk.Menu( self.menubar, tearoff=False )
        #     self.menubar.add_cascade( menu=versificationsMenu, label=_('Versifications'), underline=0 )
        #     versificationsMenu.add_command( label=_('View…'), underline=0, command=self.notWrittenYet )
        #     versificationsMenu.add_separator()
        #     versificationsMenu.add_command( label=_('Info…'), underline=0, command=self.notWrittenYet )

        #     orgsMenu = tk.Menu( self.menubar, tearoff=False )
        #     self.menubar.add_cascade( menu=orgsMenu, label=_('Bibles'), underline=0 )
        #     orgsMenu.add_command( label=_('View…'), underline=0, command=self.notWrittenYet )
        #     orgsMenu.add_separator()
        #     orgsMenu.add_command( label=_('Info…'), underline=0, command=self.notWrittenYet )

        if BibleOrgSysGlobals.debugFlag:
            debugMenu = tk.Menu( self.menubar, tearoff=False )
            self.menubar.add_cascade( menu=debugMenu, label=_('Debug'), underline=0 )
            debugMenu.add_command( label=_('View settings…'), underline=0, command=self.doViewSettings )
            debugMenu.add_separator()
            debugMenu.add_command( label=_('View log…'), underline=5, command=self.doViewLog )
            debugMenu.add_separator()
            debugMenu.add_command( label=_('Submit bug…'), underline=0, command=self.doSubmitBug )
            debugMenu.add_separator()
            debugMenu.add_command( label=_('Options…'), underline=0, command=self.notWrittenYet )

        helpMenu = tk.Menu( self.menubar, name='help', tearoff=False )
        self.menubar.add_cascade( menu=helpMenu, label=_('Help'), underline=0 )
        helpMenu.add_command( label=_('Help…'), underline=0, command=self.doHelp, accelerator=self.keyBindingDict[_('Help')][0] )
        helpMenu.add_separator()
        helpMenu.add_command( label=_('Submit bug…'), underline=0, state=tk.NORMAL if self.internetAccessEnabled else tk.DISABLED, command=self.doSubmitBug )
        helpMenu.add_separator()
        helpMenu.add_command( label=_('About…'), underline=0, command=self.doAbout, accelerator=self.keyBindingDict[_('About')][0] )
    # end of BiblelatorSettingsEditor.createNormalMenuBar

    def createTouchMenuBar( self ):
        """
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "createTouchMenuBar()" )
            assert self.touchMode

        self.createNormalMenuBar()
    # end of BiblelatorSettingsEditor.createTouchMenuBar


    def createNormalNavigationBar( self ):
        """
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "createNormalNavigationBar()" )

    # end of BiblelatorSettingsEditor.createNormalNavigationBar

    def createTouchNavigationBar( self ):
        """
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "createTouchNavigationBar()" )
            assert self.touchMode

    # end of BiblelatorSettingsEditor.createTouchNavigationBar


    def createToolBar( self ):
        """
        Create a tool bar containing several helpful buttons at the top of the main window.
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "createToolBar()" )

    # end of BiblelatorSettingsEditor.createToolBar


    def createMainButtons( self ):
        """
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "createMainButtons()" )

        xPad, yPad = (6, 8) if self.touchMode else (2, 2)

        Style().configure( 'MainButtons.TFrame', background='red' )
        Style().map("Ok.TButton", foreground=[('pressed', 'red'), ('active', 'yellow')],
                                            background=[('pressed', '!disabled', 'black'), ('active', 'pink')] )

        buttonFrame = Frame( self, cursor='hand2', relief=tk.RAISED, style='MainButtons.TFrame' )

        Button( buttonFrame, text=_("Ok"), style='Ok.TButton', command=self.doOk ) \
                        .pack( side=tk.RIGHT, padx=xPad, pady=yPad )
        Button( buttonFrame, text=_("Apply"), style='Ok.TButton', command=self.doApply ) \
                        .pack( side=tk.RIGHT, padx=xPad, pady=yPad )
        Button( buttonFrame, text=_("Cancel"), style='Ok.TButton', command=self.doCancel ) \
                        .pack( side=tk.RIGHT, padx=xPad, pady=yPad )

        buttonFrame.pack( side=tk.BOTTOM, fill=tk.X )
    # end of BiblelatorSettingsEditor.createMainButtons


    def createNotebook( self ):
        """
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "createToolBar()" )

        self.notebook = Notebook( self )

        # Adding Frames as pages for the ttk.Notebook

        # Main settings files page
        vPrint( 'Quiet', debuggingThisModule, "Create main settings files page" )
        self.settingsFilesPage = Frame( self.notebook )
        self.fdrVar = tk.StringVar()
        fdrLabel = Label( self.settingsFilesPage, text=_("Standard folder:") )
        fdrLabel.grid( row=0, column=0, padx=0, pady=2, sticky=tk.E )
        self.fdrEntry = BEntry( self.settingsFilesPage, width=50, textvariable=self.fdrVar, state=tk.DISABLED )
        self.fdrEntry.grid( row=0, column=1, padx=2, pady=2, sticky=tk.W )

        fnLabel = Label( self.settingsFilesPage, text=_("Settings name:") )
        fnLabel.grid( row=1, column=0, padx=0, pady=2, sticky=tk.E )
        self.fnVar = tk.StringVar()
        if self.fixedSettingsFlag:
            self.fnEntry = BEntry( self.settingsFilesPage, width=15, textvariable=self.fnVar, state=tk.DISABLED )
        else: # not fixed settings
            self.fnEntry = BCombobox( self.settingsFilesPage, width=15, textvariable=self.fnVar )
            #self.fnEntry['values'] = self.bookNames
            self.fnEntry.bind('<<ComboboxSelected>>', self.selectedNewSettingsFile )
            self.fnEntry.bind( '<Return>', self.selectedNewSettingsFile )
        self.fnEntry.grid( row=1, column=1, padx=2, pady=2, sticky=tk.W )


        # Main settings page
        vPrint( 'Quiet', debuggingThisModule, "Create main settings page" )
        self.mainPage = Frame( self.notebook )
        self.svVar = tk.StringVar()
        svLabel = Label( self.mainPage, text=_("Settings version:") )
        svLabel.grid( row=0, column=0, padx=0, pady=2, sticky=tk.E )
        self.svEntry = BEntry( self.mainPage, width=8, textvariable=self.svVar, state=tk.DISABLED )
        #self.fnEntry.bind( '<Return>', self.searchBBBCode )
        self.svEntry.grid( row=0, column=1, padx=2, pady=2, sticky=tk.W )
        self.pvVar = tk.StringVar()
        pvLabel = Label( self.mainPage, text=_("Program version:") )
        pvLabel.grid( row=1, column=0, padx=0, pady=2, sticky=tk.E )
        self.pvEntry = BEntry( self.mainPage, width=8, textvariable=self.pvVar, state=tk.DISABLED )
        #self.fnEntry.bind( '<Return>', self.searchBBBCode )
        self.pvEntry.grid( row=1, column=1, padx=2, pady=2, sticky=tk.W )
        self.thnVar = tk.StringVar()
        thnLabel = Label( self.mainPage, text=_("Theme name:") )
        thnLabel.grid( row=2, column=0, padx=0, pady=2, sticky=tk.E )
        self.thnEntry = BEntry( self.mainPage, width=20, textvariable=self.thnVar, state=tk.DISABLED )
        #self.fnEntry.bind( '<Return>', self.searchBBBCode )
        self.thnEntry.grid( row=2, column=1, padx=2, pady=2, sticky=tk.W )
        self.wszVar = tk.StringVar()
        wszLabel = Label( self.mainPage, text=_("Window size:") )
        wszLabel.grid( row=3, column=0, padx=0, pady=2, sticky=tk.E )
        self.wszEntry = BEntry( self.mainPage, width=12, textvariable=self.wszVar, state=tk.DISABLED )
        #self.fnEntry.bind( '<Return>', self.searchBBBCode )
        self.wszEntry.grid( row=3, column=1, padx=2, pady=2, sticky=tk.W )
        self.wposVar = tk.StringVar()
        wposLabel = Label( self.mainPage, text=_("Window position:") )
        wposLabel.grid( row=4, column=0, padx=0, pady=2, sticky=tk.E )
        self.wposEntry = BEntry( self.mainPage, width=12, textvariable=self.wposVar, state=tk.DISABLED )
        #self.fnEntry.bind( '<Return>', self.searchBBBCode )
        self.wposEntry.grid( row=4, column=1, padx=2, pady=2, sticky=tk.W )
        self.minszVar = tk.StringVar()
        minszLabel = Label( self.mainPage, text=_("Minimum size:") )
        minszLabel.grid( row=5, column=0, padx=0, pady=2, sticky=tk.E )
        self.minszEntry = BEntry( self.mainPage, width=12, textvariable=self.minszVar, state=tk.DISABLED )
        #self.fnEntry.bind( '<Return>', self.searchBBBCode )
        self.minszEntry.grid( row=5, column=1, padx=2, pady=2, sticky=tk.W )
        self.maxszVar = tk.StringVar()
        maxszLabel = Label( self.mainPage, text=_("Maximum size:") )
        maxszLabel.grid( row=6, column=0, padx=0, pady=2, sticky=tk.E )
        self.maxszEntry = BEntry( self.mainPage, width=12, textvariable=self.maxszVar, state=tk.DISABLED )
        #self.fnEntry.bind( '<Return>', self.searchBBBCode )
        self.maxszEntry.grid( row=6, column=1, padx=2, pady=2, sticky=tk.W )

        # Interface page
        vPrint( 'Quiet', debuggingThisModule, "Create interface page" )
        self.interfacePage = Frame( self.notebook )
        self.ilVar = tk.StringVar()
        ilLabel = Label( self.interfacePage, text=_("Language:") )
        ilLabel.grid( row=0, column=0, padx=0, pady=2, sticky=tk.E )
        self.ilEntry = BEntry( self.interfacePage, width=25, textvariable=self.ilVar, state=tk.DISABLED )
        #self.ilEntry.bind( '<Return>', self.searchBBBCode )
        self.ilEntry.grid( row=0, column=1, padx=2, pady=2, sticky=tk.W )
        self.icVar = tk.StringVar()
        icLabel = Label( self.interfacePage, text=_("Complexity:") )
        icLabel.grid( row=1, column=0, padx=0, pady=2, sticky=tk.E )
        self.icEntry = BEntry( self.interfacePage, width=25, textvariable=self.icVar, state=tk.DISABLED )
        #self.icEntry.bind( '<Return>', self.searchBBBCode )
        self.icEntry.grid( row=1, column=1, padx=2, pady=2, sticky=tk.W )
        self.tchVar = tk.IntVar()
        tchCb = tk.Checkbutton( self.interfacePage, text=_("Touch mode:"), variable=self.tchVar, command=self.flagChange )
        tchCb.grid( row=2, column=1, padx=0, pady=2, sticky=tk.W )
        self.tabVar = tk.IntVar()
        tabCb = tk.Checkbutton( self.interfacePage, text=_("Tablet mode:"), variable=self.tabVar, command=self.flagChange )
        tabCb.grid( row=3, column=1, padx=0, pady=2, sticky=tk.W )
        self.sdmVar = tk.IntVar()
        tabCb = tk.Checkbutton( self.interfacePage, text=_("Show debug mode:"), variable=self.sdmVar, command=self.flagChange )
        tabCb.grid( row=3, column=1, padx=0, pady=2, sticky=tk.W )

        # Internet communications page
        vPrint( 'Quiet', debuggingThisModule, "Create Internet page" )
        self.internetPage = Frame( self.notebook )
        self.iaVar = tk.IntVar()
        iaCb = tk.Checkbutton( self.internetPage, text=_("Internet access enabled"), variable=self.iaVar, command=self.flagChange )
        iaCb.grid( row=0, column=0, padx=0, pady=2, sticky=tk.W )
        self.ifVar = tk.IntVar()
        ifCb = tk.Checkbutton( self.internetPage, text=_("Internet is fast"), variable=self.ifVar, command=self.flagChange )
        ifCb.grid( row=1, column=0, padx=20, pady=2, sticky=tk.W )
        self.ieVar = tk.IntVar()
        ieCb = tk.Checkbutton( self.internetPage, text=_("Internet is expensive"), variable=self.ieVar, command=self.flagChange )
        ieCb.grid( row=2, column=0, padx=20, pady=2, sticky=tk.W )
        self.cbVar = tk.IntVar()
        cbCb = tk.Checkbutton( self.internetPage, text=_("Cloud backup enabled"), variable=self.cbVar, command=self.flagChange )
        cbCb.grid( row=3, column=0, padx=20, pady=2, sticky=tk.W )
        self.cdVar = tk.IntVar()
        cdCb = tk.Checkbutton( self.internetPage, text=_("Check for developer messages"), variable=self.cdVar, command=self.flagChange )
        cdCb.grid( row=4, column=0, padx=20, pady=2, sticky=tk.W )
        self.lmVar = tk.StringVar()
        lmLabel = Label( self.internetPage, text=_("Last message number read:") )
        lmLabel.grid( row=5, column=0, sticky=tk.E )
        self.lmEntry = BEntry( self.internetPage, width=5, textvariable=self.lmVar )
        #self.lmEntry.bind( '<Return>', self.searchBBBCode )
        self.lmEntry.grid( row=5, column=1 )
        self.usVar = tk.IntVar()
        usCb = tk.Checkbutton( self.internetPage, text=_("Send usage statistics"), variable=self.usVar, command=self.flagChange )
        usCb.grid( row=6, column=0, padx=20, pady=2, sticky=tk.W )
        self.auVar = tk.IntVar()
        auCb = tk.Checkbutton( self.internetPage, text=_("Automatic updates"), variable=self.auVar, command=self.flagChange )
        auCb.grid( row=7, column=0, padx=20, pady=2, sticky=tk.W )
        self.dvVar = tk.IntVar()
        dvCb = tk.Checkbutton( self.internetPage, text=_("Use development versions"), variable=self.dvVar, command=self.flagChange )
        dvCb.grid( row=8, column=0, padx=40, pady=2, sticky=tk.W )

        # Projects page
        vPrint( 'Quiet', debuggingThisModule, "Create projects page" )
        self.projectsPage = Frame( self.notebook )
        self.cpVar = tk.StringVar()
        cpLabel = Label( self.projectsPage, text=_("Current project name:") )
        cpLabel.grid( row=0, column=0, padx=0, pady=2, sticky=tk.E )
        self.cpEntry = BEntry( self.projectsPage, width=25, textvariable=self.cpVar )
        #self.lmEntry.bind( '<Return>', self.searchBBBCode )
        self.cpEntry.grid( row=0, column=1, padx=2, pady=2, sticky=tk.W )

        # Users page
        vPrint( 'Quiet', debuggingThisModule, "Create users page" )
        self.usersPage = Frame( self.notebook )
        self.unVar = tk.StringVar()
        unLabel = Label( self.usersPage, text=_("Current user name:") )
        unLabel.grid( row=0, column=0, padx=0, pady=2, sticky=tk.E )
        self.unEntry = BEntry( self.usersPage, width=25, textvariable=self.unVar )
        #self.lmEntry.bind( '<Return>', self.searchBBBCode )
        self.unEntry.grid( row=0, column=1, padx=2, pady=2, sticky=tk.W )
        self.uiVar = tk.StringVar()
        uiLabel = Label( self.usersPage, text=_("Current user initials:") )
        uiLabel.grid( row=1, column=0, padx=0, pady=2, sticky=tk.E )
        self.uiEntry = BEntry( self.usersPage, width=5, textvariable=self.uiVar )
        #self.lmEntry.bind( '<Return>', self.searchBBBCode )
        self.uiEntry.grid( row=1, column=1, padx=2, pady=2, sticky=tk.W )
        self.uemVar = tk.StringVar()
        uemLabel = Label( self.usersPage, text=_("User email:") )
        uemLabel.grid( row=2, column=0, padx=0, pady=2, sticky=tk.E )
        self.uemEntry = BEntry( self.usersPage, width=25, textvariable=self.uemVar )
        #self.lmEntry.bind( '<Return>', self.searchBBBCode )
        self.uemEntry.grid( row=2, column=1, padx=2, pady=2, sticky=tk.W )
        self.urVar = tk.StringVar()
        urLabel = Label( self.usersPage, text=_("User role:") )
        urLabel.grid( row=3, column=0, padx=0, pady=2, sticky=tk.E )
        self.urEntry = BEntry( self.usersPage, width=20, textvariable=self.urVar )
        #self.lmEntry.bind( '<Return>', self.searchBBBCode )
        self.urEntry.grid( row=3, column=1, padx=2, pady=2, sticky=tk.W )
        self.uasVar = tk.StringVar()
        uasLabel = Label( self.usersPage, text=_("User assignments:") )
        uasLabel.grid( row=4, column=0, padx=0, pady=2, sticky=tk.E )
        self.uasEntry = BEntry( self.usersPage, width=20, textvariable=self.uasVar )
        #self.lmEntry.bind( '<Return>', self.searchBBBCode )
        self.uasEntry.grid( row=4, column=1, padx=2, pady=2, sticky=tk.W )

        # Paths page
        vPrint( 'Quiet', debuggingThisModule, "Create paths page" )
        self.pathsPage = Frame( self.notebook )
        self.ltfVar = tk.StringVar()
        ltfLabel = Label( self.pathsPage, text=_("Last text folder:") )
        ltfLabel.grid( row=0, column=0, padx=0, pady=2, sticky=tk.E )
        self.ltfEntry = BEntry( self.pathsPage, width=35, textvariable=self.ltfVar )
        #self.lmEntry.bind( '<Return>', self.searchBBBCode )
        self.ltfEntry.grid( row=0, column=1, padx=2, pady=2, sticky=tk.W )
        self.lbfVar = tk.StringVar()
        lbfLabel = Label( self.pathsPage, text=_("Last Biblelator folder:") )
        lbfLabel.grid( row=1, column=0, padx=0, pady=2, sticky=tk.E )
        self.lbfEntry = BEntry( self.pathsPage, width=35, textvariable=self.lbfVar )
        #self.lmEntry.bind( '<Return>', self.searchBBBCode )
        self.lbfEntry.grid( row=2, column=1, padx=2, pady=2, sticky=tk.W )
        self.lpfVar = tk.StringVar()
        lpfLabel = Label( self.pathsPage, text=_("Last Paratext folder:") )
        lpfLabel.grid( row=2, column=0, padx=0, pady=2, sticky=tk.E )
        self.lpfEntry = BEntry( self.pathsPage, width=35, textvariable=self.lpfVar )
        #self.lmEntry.bind( '<Return>', self.searchBBBCode )
        self.lpfEntry.grid( row=2, column=1, padx=2, pady=2, sticky=tk.W )
        self.libfVar = tk.StringVar()
        libfLabel = Label( self.pathsPage, text=_("Last internal Bible folder:") )
        libfLabel.grid( row=3, column=0, padx=0, pady=2, sticky=tk.E )
        self.libfEntry = BEntry( self.pathsPage, width=35, textvariable=self.libfVar )
        #self.lmEntry.bind( '<Return>', self.searchBBBCode )
        self.libfEntry.grid( row=3, column=1, padx=2, pady=2, sticky=tk.W )

        # Recent files page
        vPrint( 'Quiet', debuggingThisModule, "Create recent files page" )
        self.recentFilesPage = Frame( self.notebook )
        self.rffnVars, self.rffldVars, self.rftypVars = [], [], []
        for rr in range( MAX_RECENT_FILES ):
            self.rffnVars.append( tk.StringVar() ); self.rffldVars.append( tk.StringVar() ); self.rftypVars.append( tk.StringVar() );
            Label( self.recentFilesPage, text='{}:'.format(rr+1) ).grid( row=2*rr, column=0, padx=0, pady=3, sticky=tk.E )
            BEntry( self.recentFilesPage, width=30, textvariable=self.rffnVars[rr] ).grid( row=2*rr, column=1, padx=2, pady=3, sticky=tk.W )
            BEntry( self.recentFilesPage, width=30, textvariable=self.rftypVars[rr] ).grid( row=2*rr, column=2, padx=2, pady=3, sticky=tk.W )
            BEntry( self.recentFilesPage, width=60, textvariable=self.rffldVars[rr] ).grid( row=2*rr+1, column=1, columnspan=2, padx=2, pady=1, sticky=tk.W )

        # Bible BCV (book/chapter/verse) page
        vPrint( 'Quiet', debuggingThisModule, "Create BCV page" )
        self.BCVGroupsPage = Frame( self.notebook )
        self.gBOSVar = tk.StringVar()
        gBOSLabel = Label( self.BCVGroupsPage, text=_("Generic BOS name:") )
        gBOSLabel.grid( row=0, column=0, padx=0, pady=2, sticky=tk.E )
        self.gBOSEntry = BEntry( self.BCVGroupsPage, width=35, textvariable=self.gBOSVar )
        #self.lmEntry.bind( '<Return>', self.searchBBBCode )
        self.gBOSEntry.grid( row=0, column=1, padx=2, pady=2, sticky=tk.W )
        self.cgVar = tk.StringVar()
        cgLabel = Label( self.BCVGroupsPage, text=_("Current group:") )
        cgLabel.grid( row=1, column=0, padx=0, pady=2, sticky=tk.E )
        self.cgEntry = BEntry( self.BCVGroupsPage, width=3, textvariable=self.cgVar )
        #self.lmEntry.bind( '<Return>', self.searchBBBCode )
        self.cgEntry.grid( row=1, column=1, padx=2, pady=2, sticky=tk.W )
        self.gaVar = tk.StringVar()
        gaLabel = Label( self.BCVGroupsPage, text=_("Group A:") )
        gaLabel.grid( row=2, column=0, padx=0, pady=2, sticky=tk.E )
        self.gaEntry = BEntry( self.BCVGroupsPage, width=12, textvariable=self.gaVar )
        #self.lmEntry.bind( '<Return>', self.searchBBBCode )
        self.gaEntry.grid( row=2, column=1, padx=2, pady=2, sticky=tk.W )
        self.gbVar = tk.StringVar()
        gbLabel = Label( self.BCVGroupsPage, text=_("Group B:") )
        gbLabel.grid( row=3, column=0, padx=0, pady=2, sticky=tk.E )
        self.gbEntry = BEntry( self.BCVGroupsPage, width=12, textvariable=self.gbVar )
        #self.lmEntry.bind( '<Return>', self.searchBBBCode )
        self.gbEntry.grid( row=3, column=1, padx=2, pady=2, sticky=tk.W )
        self.gcVar = tk.StringVar()
        gcLabel = Label( self.BCVGroupsPage, text=_("Group C:") )
        gcLabel.grid( row=4, column=0, padx=0, pady=2, sticky=tk.E )
        self.gcEntry = BEntry( self.BCVGroupsPage, width=12, textvariable=self.gcVar )
        #self.lmEntry.bind( '<Return>', self.searchBBBCode )
        self.gcEntry.grid( row=4, column=1, padx=2, pady=2, sticky=tk.W )
        self.gdVar = tk.StringVar()
        gdLabel = Label( self.BCVGroupsPage, text=_("Group D:") )
        gdLabel.grid( row=5, column=0, padx=0, pady=2, sticky=tk.E )
        self.gdEntry = BEntry( self.BCVGroupsPage, width=12, textvariable=self.gdVar )
        #self.lmEntry.bind( '<Return>', self.searchBBBCode )
        self.gdEntry.grid( row=5, column=1, padx=2, pady=2, sticky=tk.W )

        # Current windows page
        vPrint( 'Quiet', debuggingThisModule, "Create current windows page" )
        self.currentWindowsPage = Frame( self.notebook )
        if __name__ == '__main__':
            pass
        else:
            self.currentWindowsTextBox = ScrolledText( self.currentWindowsPage, bg='orange' )
            self.currentWindowsTextBox.tag_configure( 'emp', font='helvetica 10 bold' )
            #self.currentWindowsTextBox.insert( tk.END, 'Organisations' )
            self.currentWindowsTextBox.grid( row=0, column=4, rowspan=2, sticky=tk.N+tk.S+tk.E )
            self.currentWindowsTextBox.insert( tk.END, "We cannot adjust the current windows from inside Biblelator.\n\nIf you wish to adjust current windows, please close Biblelator and run BiblelatorSettingsEditor.py in stand-alone mode." )

        vPrint( 'Quiet', debuggingThisModule, "Add all pages" )
        self.notebook.add( self.settingsFilesPage, text=_("Settings files") )
        self.notebook.add( self.mainPage, text=_("Main") )
        self.notebook.add( self.interfacePage, text=_("Interface") )
        self.notebook.add( self.internetPage, text=_("Internet") )
        self.notebook.add( self.projectsPage, text=_("Projects") )
        self.notebook.add( self.usersPage, text=_("Users") )
        self.notebook.add( self.pathsPage, text=_("Paths") )
        self.notebook.add( self.recentFilesPage, text=_("Recent files") )
        self.notebook.add( self.BCVGroupsPage, text=_("BCV groups") )
        self.notebook.add( self.currentWindowsPage, text=_("Current windows") )
        self.notebook.pack( expand=tk.YES, fill=tk.BOTH )

        self.loadSettingsIntoTabs()
        self.somethingChanged = False
    # end of BiblelatorSettingsEditor.createNotebook


    def flagChange( self ): self.somethingChanged = True


    def loadSettingsIntoTabs( self ):
        """
        Load the current settings for self.INIname into self.settings.data
            and then load them into the variables for our editor.
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "loadSettingsIntoTabs() for {!r}".format( self.INIname ) )

        self.settings = ApplicationSettings( self.homeFolderPath, DATA_FOLDER_NAME, SETTINGS_SUBFOLDER_NAME, self.INIname )
        self.settings.load()
        self.settingsChangedFlag = False

        self.fdrVar.set( self.settings.settingsFolder )
        self.fnVar.set( self.INIname )
        if not self.fixedSettingsFlag:
            self.iniList = []
            for something in os.listdir( self.settings.settingsFolder ):
                somepath = os.path.join( self.settings.settingsFolder, something )
                if os.path.isfile( somepath ):
                    if somepath.upper().endswith( '.INI' ):
                        self.iniList.append( something[:-4] )
            self.fnEntry['values'] = self.iniList

        try: self.svVar.set( self.settings.data[MAIN_APP_NAME]['settingsVersion'] )
        except KeyError: self.svVar.set( '' )
        try: self.pvVar.set( self.settings.data[MAIN_APP_NAME]['PROGRAM_VERSION'] )
        except KeyError: self.pvVar.set( '' )
        try: self.thnVar.set( self.settings.data['Interface']['themeName'] )
        except KeyError: self.thnVar.set( 'default' )
        try: self.wszVar.set( self.settings.data[MAIN_APP_NAME]['windowSize'] )
        except KeyError: self.wszVar.set( '' )
        try: self.wposVar.set( self.settings.data[MAIN_APP_NAME]['windowPosition'] )
        except KeyError: self.wposVar.set( '' )
        try: self.minszVar.set( self.settings.data[MAIN_APP_NAME]['minimumSize'] )
        except KeyError: self.minszVar.set( '' )
        try: self.maxszVar.set( self.settings.data[MAIN_APP_NAME]['maximumSize'] )
        except KeyError: self.maxszVar.set( '' )

        try: self.ilVar.set( self.settings.data['Interface']['interfaceLanguage'] )
        except KeyError: self.ilVar.set( 'Default' )
        try: self.icVar.set( self.settings.data['Interface']['interfaceComplexity'] )
        except KeyError: self.icVar.set( 'Default' )
        try: self.tchVar.set( self.settings.data['Interface']['touchMode'] == 'True' )
        except KeyError: self.tchVar.set( False )
        try: self.tabVar.set( self.settings.data['Interface']['tabletMode'] == 'True' )
        except KeyError: self.tabVar.set( False )
        try: self.sdmVar.set( self.settings.data['Interface']['showDebugMenu'] == 'True' )
        except KeyError: self.sdmVar.set( False )

        try: self.iaVar.set( self.settings.data['Internet']['internetAccess'] == 'Enabled' )
        except KeyError: self.iaVar.set( True )
        try: self.ifVar.set( self.settings.data['Internet']['internetFast'].lower() in ('true' ,'yes',) )
        except KeyError: self.ifVar.set( True )
        try: self.ieVar.set( self.settings.data['Internet']['internetExpensive'].lower() in ('true' ,'yes',) )
        except KeyError: self.ieVar.set( True )
        try: self.cbVar.set( self.settings.data['Internet']['cloudBackups'] == 'Enabled' )
        except KeyError: self.cbVar.set( True )
        try: self.cdVar.set( self.settings.data['Internet']['checkForDeveloperMessages'] == 'Enabled' )
        except KeyError: self.cdVar.set( True )
        try: self.lmVar.set( self.settings.data['Internet']['lastMessageNumberRead'] )
        except KeyError: self.lmVar.set( 0 )
        try: self.usVar.set( self.settings.data['Internet']['sendUsageStatistics'] == 'Enabled' )
        except KeyError: self.suVar.set( True )
        try: self.auVar.set( self.settings.data['Internet']['automaticUpdates'] == 'Enabled' )
        except KeyError: self.auVar.set( True )
        try: self.dvVar.set( self.settings.data['Internet']['useDevelopmentVersions'] == 'Enabled' )
        except KeyError: self.dvVar.set( False )

        try: self.cpVar.set( self.settings.data['Project']['currentProjectName'] )
        except KeyError: self.cpVar.set( '' )

        try: self.unVar.set( self.settings.data['Users']['currentUserName'] )
        except KeyError: self.unVar.set( '' )
        try: self.uiVar.set( self.settings.data['Users']['currentUserInitials'] )
        except KeyError: self.uiVar.set( '' )
        try: self.uemVar.set( self.settings.data['Users']['currentUserEmail'] )
        except KeyError: self.uemVar.set( '' )
        try: self.urVar.set( self.settings.data['Users']['currentUserRole'] )
        except KeyError: self.urVar.set( 'Translator' )
        try: self.uasVar.set( self.settings.data['Users']['currentUserAssignments'] )
        except KeyError: self.uasVar.set( 'ALL' )

        try: self.ltfVar.set( self.settings.data['Paths']['lastFileDir'] )
        except KeyError: self.ltfVar.set( '' )
        try: self.lbfVar.set( self.settings.data['Paths']['lastBiblelatorFileDir'] )
        except KeyError: self.lbfVar.set( '' )
        try: self.lpfVar.set( self.settings.data['Paths']['lastParatextFileDir'] )
        except KeyError: self.lpfVar.set( '' )
        try: self.libfVar.set( self.settings.data['Paths']['lastInternalBibleDir'] )
        except KeyError: self.libfVar.set( '' )

        for rr in range( MAX_RECENT_FILES ):
            try:
                self.rffnVars[rr].set( self.settings.data['RecentFiles']['recent{}Filename'.format(rr+1)] )
                self.rffldVars[rr].set( self.settings.data['RecentFiles']['recent{}Folder'.format(rr+1)] )
                self.rftypVars[rr].set( self.settings.data['RecentFiles']['recent{}Type'.format(rr+1)] )
            except KeyError: self.rffnVars[rr].set( '' ); self.rffldVars[rr].set( '' ); self.rftypVars[rr].set( '' )

        try: self.gBOSVar.set( self.settings.data['BCVGroups']['genericBibleOrganisationalSystemName'] )
        except KeyError: self.gBOSVar.set( '' )
        try: self.cgVar.set( self.settings.data['BCVGroups']['currentGroup'] )
        except KeyError: self.cgVar.set( 'A' )
        try: self.gaVar.set( '{} {}:{}'.format( self.settings.data['BCVGroups']['A-Book'], self.settings.data['BCVGroups']['A-Chapter'], self.settings.data['BCVGroups']['A-Verse'] ) )
        except KeyError: self.gaVar.set( 'GEN 1:1' )
        try: self.gbVar.set( '{} {}:{}'.format( self.settings.data['BCVGroups']['B-Book'], self.settings.data['BCVGroups']['B-Chapter'], self.settings.data['BCVGroups']['B-Verse'] ) )
        except KeyError: self.gbVar.set( 'GEN 1:1' )
        try: self.gcVar.set( '{} {}:{}'.format( self.settings.data['BCVGroups']['C-Book'], self.settings.data['BCVGroups']['C-Chapter'], self.settings.data['BCVGroups']['C-Verse'] ) )
        except KeyError: self.gcVar.set( 'GEN 1:1' )
        try: self.gdVar.set( '{} {}:{}'.format( self.settings.data['BCVGroups']['D-Book'], self.settings.data['BCVGroups']['D-Chapter'], self.settings.data['BCVGroups']['D-Verse'] ) )
        except KeyError: self.gdVar.set( 'GEN 1:1' )
    # end of BiblelatorSettingsEditor.loadSettingsIntoTabs


    def updateSettingsFromTabs( self ):
        """
        Update the settings from the editor, and return True/False if they have changed.
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "updateSettingsFromTabs()" )

        changed = False

        last = self.settings.data['Internet']['internetAccess']
        now = 'Enabled' if self.iaVar.get() else 'Disabled'
        self.settings.data['Internet']['internetAccess'] = now
        if now!=last: changed = True

        if changed: assert self.somethingChanged
        return changed
    # end of BiblelatorSettingsEditor.updateSettingsFromTabs


    def halt( self ):
        """
        Halts the program immediately without saving any files or settings.
        Only used in debug mode.
        """
        logging.critical( "User selected HALT in DEBUG MODE. Not saving any files or settings!" )
        self.quit()
    # end of BiblelatorSettingsEditor.halt


    def createDebugToolBar( self ):
        """
        Create a debug tool bar containing several additional buttons at the top of the main window.
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "createDebugToolBar()" )

        xPad, yPad = (6, 8) if self.touchMode else (2, 2)

        Style().configure( 'DebugToolBar.TFrame', background='red' )
        Style().map("Halt.TButton", foreground=[('pressed', 'red'), ('active', 'yellow')],
                                            background=[('pressed', '!disabled', 'black'), ('active', 'pink')] )

        toolbar = Frame( self, cursor='hand2', relief=tk.RAISED, style='DebugToolBar.TFrame' )
        Button( toolbar, text='Halt', style='Halt.TButton', command=self.halt ) \
                        .pack( side=tk.RIGHT, padx=xPad, pady=yPad )
        Button( toolbar, text='Save settings', command=lambda: writeSettingsFile(self) ) \
                        .pack( side=tk.RIGHT, padx=xPad, pady=yPad )
        toolbar.pack( side=tk.TOP, fill=tk.X )
    # end of BiblelatorSettingsEditor.createDebugToolBar


    def createStatusBar( self ):
        """
        Create a status bar containing only one text label at the bottom of the main window.
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "createStatusBar()" )

        #Style().configure( 'StatusBar.TLabel', background='pink' )
        #Style().configure( 'StatusBar.TLabel', background='DarkOrange1' )
        Style().configure( 'StatusBar.TLabel', background='forest green' )

        self.statusTextVariable = tk.StringVar()
        self.statusTextLabel = Label( self.rootWindow, relief=tk.SUNKEN,
                                    textvariable=self.statusTextVariable, style='StatusBar.TLabel' )
                                    #, font=('arial',16,tk.NORMAL) )
        self.statusTextLabel.pack( side=tk.BOTTOM, fill=tk.X )
        self.statusTextVariable.set( '' ) # first initial value
        self.setWaitStatus( _("Starting up…") )
    # end of BiblelatorSettingsEditor.createStatusBar


    def createMainKeyboardBindings( self ):
        """
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "createMainKeyboardBindings()" )

        self.myKeyboardBindingsList = []
        for name,command in ( ('Help',self.doHelp), ('About',self.doAbout), ('Quit',self.doCloseMe) ):
            if name in self.keyBindingDict:
                for keyCode in self.keyBindingDict[name][1:]:
                    #vPrint( 'Quiet', debuggingThisModule, "Bind {} for {}".format( repr(keyCode), repr(name) ) )
                    self.rootWindow.bind( keyCode, command )
                self.myKeyboardBindingsList.append( (name,self.keyBindingDict[name][0],) )
            else: logging.critical( 'No key binding available for {!r}'.format( name ) )

        # These bindings apply to/from all windows
        #self.bind_all( '<Alt-Up>', self.doGotoPreviousVerse )
        #self.bind_all( '<Alt-Down>', self.doGotoNextVerse )
        #self.bind_all( '<Alt-comma>', self.doGotoPreviousChapter )
        #self.bind_all( '<Alt-period>', self.doGotoNextChapter )
        #self.bind_all( '<Alt-bracketleft>', self.doGotoPreviousBook )
        #self.bind_all( '<Alt-bracketright>', self.doGotoNextBook )
    # end of BiblelatorSettingsEditor.createMainKeyboardBindings()


    #def addRecentFile( self, threeTuple ):
        #"""
        #Puts most recent first
        #"""
        #if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            #vPrint( 'Quiet', debuggingThisModule, "addRecentFile( {} )".format( threeTuple ) )
            #assert len(threeTuple) == 3

        #try: self.recentFiles.remove( threeTuple ) # Remove a duplicate if present
        #except ValueError: pass
        #self.recentFiles.insert( 0, threeTuple ) # Put this one at the beginning of the lis
        #if len(self.recentFiles)>MAX_RECENT_FILES: self.recentFiles.pop() # Remove the last one if necessary
        #self.createNormalMenuBar()
    ## end of BiblelatorSettingsEditor.addRecentFile()


    def notWrittenYet( self ):
        errorBeep()
        showError( self, _("Not implemented"), _("Not yet available, sorry") )
    # end of BiblelatorSettingsEditor.notWrittenYet


    def setStatus( self, newStatusText='' ):
        """
        Set (or clear) the status bar text.
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "setStatus( {!r} )".format( newStatusText ) )

        #vPrint( 'Quiet', debuggingThisModule, "SB is", repr( self.statusTextVariable.get() ) )
        if newStatusText != self.statusTextVariable.get(): # it's changed
            #self.statusBarTextWidget.configure( state=tk.NORMAL )
            #self.statusBarTextWidget.delete( tkSTART, tk.END )
            #if newStatusText:
                #self.statusBarTextWidget.insert( tkSTART, newStatusText )
            #self.statusBarTextWidget.configure( state=tk.DISABLED ) # Don't allow editing
            #self.statusText = newStatusText
            Style().configure( 'StatusBar.TLabel', foreground='white', background='purple' )
            self.statusTextVariable.set( newStatusText )
            self.statusTextLabel.update()
    # end of BiblelatorSettingsEditor.setStatus

    def setErrorStatus( self, newStatusText ):
        """
        Set the status bar text and change the cursor to the wait/hourglass cursor.
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "setErrorStatus( {!r} )".format( newStatusText ) )

        #self.rootWindow.configure( cursor='watch' ) # 'wait' can only be used on Windows
        #self.statusTextLabel.configure( style='StatusBar.TLabelWait' )
        self.setStatus( newStatusText )
        Style().configure( 'StatusBar.TLabel', foreground='yellow', background='red' )
        self.update()
    # end of BiblelatorSettingsEditor.setErrorStatus

    def setWaitStatus( self, newStatusText ):
        """
        Set the status bar text and change the cursor to the wait/hourglass cursor.
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "setWaitStatus( {!r} )".format( newStatusText ) )

        self.rootWindow.configure( cursor='watch' ) # 'wait' can only be used on Windows
        #self.statusTextLabel.configure( style='StatusBar.TLabelWait' )
        self.setStatus( newStatusText )
        Style().configure( 'StatusBar.TLabel', foreground='black', background='DarkOrange1' )
        self.update()
    # end of BiblelatorSettingsEditor.setWaitStatus

    def setReadyStatus( self ):
        """
        Sets the status line to "Ready"
            and sets the cursor to the normal cursor
        unless we're still starting
            (this covers any slow start-up functions that don't yet set helpful statuses)
        """
        if self.starting: self.setWaitStatus( _("Starting up…") )
        else: # we really are ready
            #self.statusTextLabel.configure( style='StatusBar.TLabelReady' )
            self.setStatus( _("Ready") )
            Style().configure( 'StatusBar.TLabel', foreground='yellow', background='forest green' )
            self.configure( cursor='' )
    # end of BiblelatorSettingsEditor.setReadyStatus


    def setDebugText( self, newMessage=None ):
        """
        """
        if debuggingThisModule:
            #vPrint( 'Quiet', debuggingThisModule, "setDebugText( {!r} )".format( newMessage ) )
            assert BibleOrgSysGlobals.debugFlag

        logging.info( 'Debug: ' + newMessage ) # Not sure why logging.debug isn't going into the file! XXXXXXXXXXXXX
        self.debugTextBox.configure( state=tk.NORMAL ) # Allow editing
        self.debugTextBox.delete( tkSTART, tk.END ) # Clear everything
        self.debugTextBox.insert( tk.END, 'DEBUGGING INFORMATION:' )
        if self.lastDebugMessage: self.debugTextBox.insert( tk.END, '\nWas: ' + self.lastDebugMessage )
        if newMessage:
            self.debugTextBox.insert( tk.END, '\n' )
            self.debugTextBox.insert( tk.END, 'Msg: ' + newMessage, 'emp' )
            self.lastDebugMessage = newMessage
        self.debugTextBox.insert( tk.END, '\n\n{} child windows:'.format( len(self.childWindows) ) )
        for j, appWin in enumerate( self.childWindows ):
            #try: extra = ' ({})'.format( appWin.BCVUpdateType )
            #except AttributeError: extra = ''
            self.debugTextBox.insert( tk.END, "\n  {} wT={} gWT={} {} modID={} cVM={} fVM={} BCV={}" \
                                    .format( j+1,
                                        appWin.windowType,
                                        #appWin.windowType.replace('ChildWindow',''),
                                        appWin.genericWindowType,
                                        #appWin.genericWindowType.replace('Resource',''),
                                        appWin.winfo_geometry(), appWin.moduleID,
                                        appWin._contextViewMode if 'Bible' in appWin.genericWindowType else 'N/A',
                                        appWin._formatViewMode if 'Bible' in appWin.genericWindowType else 'N/A',
                                        appWin.BCVUpdateType if 'Bible' in appWin.genericWindowType else 'N/A' ) )
                                        #extra ) )
        #self.debugTextBox.insert( tk.END, '\n{} resource frames:'.format( len(self.childWindows) ) )
        #for j, projFrame in enumerate( self.childWindows ):
            #self.debugTextBox.insert( tk.END, "\n  {} {}".format( j, projFrame ) )
        self.debugTextBox.configure( state=tk.DISABLED ) # Don't allow editing
    # end of BiblelatorSettingsEditor.setDebugText


    def doChangeTheme( self, newThemeName ):
        """
        Set the window theme to the given scheme.
        """
        if BibleOrgSysGlobals.debugFlag:
            vPrint( 'Quiet', debuggingThisModule, "doChangeTheme( {!r} )".format( newThemeName ) )
            assert newThemeName
            if debuggingThisModule: self.setDebugText( 'Set theme to {!r}'.format( newThemeName ) )

        self.themeName = newThemeName
        try:
            self.style.theme_use( newThemeName )
        except tk.TclError as err:
            showError( self, 'Error', err )
    # end of BiblelatorSettingsEditor.doChangeTheme


    def selectedNewSettingsFile( self, event=None ):
        """
        Handle a new settings files selected from the GUI dropbox.
        """
        enteredFilename = self.fnVar.get()
        self.logUsage( PROGRAM_NAME, debuggingThisModule, 'selectedNewSettingsFile' )
        if 1 or BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "selectedNewSettingsFile( {} ) for {!r}".format( event, enteredFilename ) )
            #vPrint( 'Quiet', debuggingThisModule, dir(event) )

        self.doApply() # Save any changes to current settings file
        self.INIname = enteredFilename
        self.loadSettingsIntoTabs()
    # end of BiblelatorSettingsEditor.selectedNewSettingsFile


    def doOk( self, event=None ):
        """
        Pop-up dialog
        """
        if 1 or BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "BiblelatorSettingsEditor.doOk( {} )".format( event ) )

        if self.settingsChangedFlag: halt
        self.doCloseMe()
    # end of BiblelatorSettingsEditor.doOk

    def doApply( self, event=None ):
        """
        Pop-up dialog
        """
        if 1 or BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "BiblelatorSettingsEditor.doApply( {} )".format( event ) )

        if self.settingsChangedFlag: halt
    # end of BiblelatorSettingsEditor.doApply

    def doCancel( self, event=None ):
        """
        Pop-up dialog
        """
        if 1 or BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "BiblelatorSettingsEditor.doCancel( {} )".format( event ) )

        if self.settingsChangedFlag: halt
        self.doCloseMe()
    # end of BiblelatorSettingsEditor.doCancel


    def doViewSettings( self ):
        """
        Open a pop-up text window with the current settings displayed.
        """
        viewSettings( self )
        #if BibleOrgSysGlobals.debugFlag:
            #vPrint( 'Never', debuggingThisModule, "doViewSettings()" )
            #self.setDebugText( "doViewSettings…" )
        #tEW = TextEditWindow( self )
        ##if windowGeometry: tEW.geometry( windowGeometry )
        #if not tEW.setFilepath( self.settings.settingsFilepath ) \
        #or not tEW.loadText():
            #tEW.doClose()
            #showError( self, SHORT_PROGRAM_NAME, _("Sorry, unable to open settings file") )
            #if BibleOrgSysGlobals.debugFlag and debuggingThisModule: self.setDebugText( "Failed doViewSettings" )
        #else:
            #self.childWindows.append( tEW )
            #if BibleOrgSysGlobals.debugFlag and debuggingThisModule: self.setDebugText( "Finished doViewSettings" )
        #self.setReadyStatus()
    # end of BiblelatorSettingsEditor.doViewSettings


    def doViewLog( self ):
        """
        Open a pop-up text window with the current log displayed.
        """
        vPrint( 'Never', debuggingThisModule, "doViewLog()" )
        if debuggingThisModule: self.setDebugText( "doViewLog…" )

        self.setWaitStatus( _("doViewLog…") )
        filename = PROGRAM_NAME.replace('/','-').replace(':','_').replace('\\','_') + '_log.txt'
        tEW = TextEditWindow( self )
        #if windowGeometry: tEW.geometry( windowGeometry )
        if not tEW.setPathAndFile( self.loggingFolderPath, filename ) \
        or not tEW.loadText():
            tEW.doClose()
            showError( self, SHORT_PROGRAM_NAME, _("Sorry, unable to open log file") )
            if BibleOrgSysGlobals.debugFlag: self.setDebugText( "Failed doViewLog" )
        else:
            self.childWindows.append( tEW )
            #if BibleOrgSysGlobals.debugFlag: self.setDebugText( "Finished doViewLog" ) # Don't do this -- adds to the log immediately
        self.setReadyStatus()
    # end of BiblelatorSettingsEditor.doViewLog


    def doGotoInfo( self, event=None ):
        """
        Pop-up dialog giving goto/reference info.
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "BiblelatorSettingsEditor.doGotoInfo( {} )".format( event ) )

        infoString = 'Current location:\n' \
                 + '\nBible Organisational System (BOS):\n' \
                 + '  Name: {}\n'.format( self.genericBibleOrganisationalSystem.getOrganisationalSystemName() ) \
                 + '  Versification: {}\n'.format( self.genericBibleOrganisationalSystem.getOrganisationalSystemValue( 'versificationSystem' ) ) \
                 + '  Book Order: {}\n'.format( self.genericBibleOrganisationalSystem.getOrganisationalSystemValue( 'bookOrderSystem' ) ) \
                 + '  Book Names: {}\n'.format( self.genericBibleOrganisationalSystem.getOrganisationalSystemValue( 'punctuationSystem' ) ) \
                 + '  Books: {}'.format( self.genericBibleOrganisationalSystem.getBookList() )
        showInfo( self, 'Goto Information', infoString )
    # end of BiblelatorSettingsEditor.doGotoInfo


    def logUsage( self, p1, p2, p3 ):
        """
        Not required in this app.
        """
        pass
    # end of BiblelatorSettingsEditor.logUsage


    def doHelp( self, event=None ):
        """
        Display a help box.
        """
        from Biblelator.Dialogs.Help import HelpBox
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "doHelp()" )

        helpInfo = programNameVersion
        helpInfo += "\n\nBasic instructions:"
        helpInfo += "\n  Click on a tab to view that subset of the Biblelator settings."
        helpInfo += "\n\nNOTE: It is also possible to edit the settings file(s) directly, but the editor should give you more context help with the various options."
        helpInfo += "\n\nKeyboard shortcuts:"
        #for name,shortcut in self.myKeyboardBindingsList:
            #helpInfo += "\n  {}\t{}".format( name, shortcut )
        #helpInfo += "\n\n  {}\t{}".format( 'Prev Verse', 'Alt+UpArrow' )
        #helpInfo += "\n  {}\t{}".format( 'Next Verse', 'Alt+DownArrow' )
        #helpInfo += "\n  {}\t{}".format( 'Prev Chapter', 'Alt+, (<)' )
        #helpInfo += "\n  {}\t{}".format( 'Next Chapter', 'Alt+. (>)' )
        #helpInfo += "\n  {}\t{}".format( 'Prev Book', 'Alt+[' )
        #helpInfo += "\n  {}\t{}".format( 'Next Book', 'Alt+]' )
        helpImage = DATAFILES_FOLDERPATH.joinpath( 'BiblelatorLogoSmall.gif' )
        hb = HelpBox( self.rootWindow, SHORT_PROGRAM_NAME, helpInfo, helpImage )
    # end of BiblelatorSettingsEditor.doHelp


    def doSubmitBug( self, event=None ):
        """
        Prompt the user to enter a bug report,
            collect other useful settings, etc.,
            and then send it all somewhere.
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "doSubmitBug()" )

        if not self.internetAccessEnabled: # we need to warn
            showError( self, SHORT_PROGRAM_NAME, 'You need to allow Internet access first!' )
            return

        from Biblelator.Dialogs.About import AboutBox

        submitInfo = programNameVersion
        submitInfo += "\n  This program is not yet finished but we'll add this eventually!"
        ab = AboutBox( self.rootWindow, SHORT_PROGRAM_NAME, submitInfo )
    # end of BiblelatorSettingsEditor.doSubmitBug


    def doAbout( self, event=None ):
        """
        Display an about box.
        """
        from Biblelator.Dialogs.About import AboutBox
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "doAbout()" )

        aboutInfo = programNameVersion
        aboutInfo += "\nAn editor for the Biblelator (Bible translation editor) settings." \
            + "\n\nThis is still an unfinished alpha test version, but it should allow you to view (not yet alter/save) various settings in Biblelator." \
            + "\n\n{} is written in Python. For more information see our web page at Freely-Given.org/Software/Biblelator".format( SHORT_PROGRAM_NAME )
        aboutImage = DATAFILES_FOLDERPATH.joinpath( 'BiblelatorLogoSmall.gif' )
        ab = AboutBox( self.rootWindow, SHORT_PROGRAM_NAME, aboutInfo, aboutImage )
    # end of BiblelatorSettingsEditor.doAbout


    #def doProjectClose( self ):
        #"""
        #"""
        #if BibleOrgSysGlobals.debugFlag and debuggingThisModule: vPrint( 'Quiet', debuggingThisModule, "doProjectClose()" )
        #self.notWrittenYet()
    ## end of BiblelatorSettingsEditor.doProjectClose


    #def doWriteSettingsFile( self ):
        #"""
        #Update our program settings and save them.
        #"""
        #writeSettingsFile( self )
    ### end of BiblelatorSettingsEditor.writeSettingsFile


    def doCloseMyChildWindows( self ):
        """
        Save files first, and then close child windows.
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "BiblelatorSettingsEditor.doCloseMyChildWindows()" )

        # Try to close edit windows first coz they might have work to save
        for appWin in self.childWindows.copy():
            if 'Editor' in appWin.genericWindowType and appWin.modified():
                appWin.doClose()
                #appWin.onCloseEditor( terminate=False )
                ##if appWin.saveChangesAutomatically: appWin.doSave( 'Auto from app close' )
                ##else: appWin.onCloseEditor()

        # See if they saved/closed them all
        haveModifications = False
        for appWin in self.childWindows:
            if 'Editor' in appWin.genericWindowType and appWin.modified():
                if appWin.modified(): # still???
                    haveModifications = True; break
        if haveModifications:
            showError( self, _("Save files"), _("You need to save or close your work first.") )
            return False

        # Should be able to close all apps now
        for appWin in self.childWindows.copy():
            appWin.doClose()
        return True
    # end of BiblelatorSettingsEditor.doCloseMyChildWindows


    def doCloseMe( self ):
        """
        Save files first, and then end the application.
        """
        if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            vPrint( 'Quiet', debuggingThisModule, "BiblelatorSettingsEditor.doCloseMe()" )
        elif BibleOrgSysGlobals.verbosityLevel > 0:
            vPrint( 'Quiet', debuggingThisModule, _("{} is closing down…").format( SHORT_PROGRAM_NAME ) )

        #writeSettingsFile( self )
        if self.doCloseMyChildWindows():
            self.rootWindow.destroy()
    # end of BiblelatorSettingsEditor.doCloseMe
# end of class BiblelatorSettingsEditor



def openBiblelatorSettingsEditor( parent ):
    """
    Open the BOS Manager as a child window.

    This is used when the BOS Manager is used inside another program.
    """
    if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
        vPrint( 'Quiet', debuggingThisModule, "BiblelatorSettingsEditor.openBiblelatorSettingsEditor( {} )".format( parent ) )

    myWin = tk.Toplevel( parent )
    application = BiblelatorSettingsEditor( myWin, parent.homeFolderPath, parent.loggingFolderPath, parent.iconImage )
# end of BiblelatorSettingsEditor.openBiblelatorSettingsEditor



def briefDemo() -> None:
    """
    Unattended demo program to handle command line parameters and then run what they want.

    Which windows open depends on the saved settings from the last use.
    """
    BibleOrgSysGlobals.introduceProgram( __name__, programNameVersion, LAST_MODIFIED_DATE )

    tkRootWindow = tk.Tk()
    if BibleOrgSysGlobals.debugFlag:
        vPrint( 'Quiet', debuggingThisModule, 'Windowing system is', repr( tkRootWindow.tk.call('tk', 'windowingsystem') ) )
    tkRootWindow.title( programNameVersion )

    # Set the window icon and title
    iconImage = tk.PhotoImage( file=DATAFILES_FOLDERPATH.joinpath( 'Biblelator.gif' ) )
    tkRootWindow.tk.call( 'wm', 'iconphoto', tkRootWindow._w, iconImage )
    tkRootWindow.title( programNameVersion + ' ' + _('starting') + '…' )

    homeFolderPath = BibleOrgSysGlobals.findHomeFolderPath()
    loggingFolderPath = os.path.join( homeFolderPath, DATA_FOLDER_NAME, LOGGING_SUBFOLDER_NAME )
    settings = ApplicationSettings( homeFolderPath, DATA_FOLDER_NAME, SETTINGS_SUBFOLDER_NAME, PROGRAM_NAME )
    settings.load()

    application = BiblelatorSettingsEditor( tkRootWindow, homeFolderPath, loggingFolderPath, iconImage )
    # Calls to the window manager class (wm in Tk)
    #application.master.title( programNameVersion )
    #application.master.minsize( application.minimumXSize, application.minimumYSize )

    # Program a shutdown
    tkRootWindow.after( 2_000, tkRootWindow.destroy ) # Destroy the widget after 2 seconds

    # Start the program running
    tkRootWindow.mainloop()
# end of BiblelatorSettingsEditor.briefDemo

def fullDemo() -> None:
    """
    Full demo to check class is working
    """
    BibleOrgSysGlobals.introduceProgram( __name__, programNameVersion, LAST_MODIFIED_DATE )

    tkRootWindow = tk.Tk()
    if BibleOrgSysGlobals.debugFlag:
        vPrint( 'Quiet', debuggingThisModule, 'Windowing system is', repr( tkRootWindow.tk.call('tk', 'windowingsystem') ) )
    tkRootWindow.title( programNameVersion )

    # Set the window icon and title
    iconImage = tk.PhotoImage( file=DATAFILES_FOLDERPATH.joinpath( 'Biblelator.gif' ) )
    tkRootWindow.tk.call( 'wm', 'iconphoto', tkRootWindow._w, iconImage )
    tkRootWindow.title( programNameVersion + ' ' + _('starting') + '…' )

    homeFolderPath = BibleOrgSysGlobals.findHomeFolderPath()
    loggingFolderPath = os.path.join( homeFolderPath, DATA_FOLDER_NAME, LOGGING_SUBFOLDER_NAME )
    settings = ApplicationSettings( homeFolderPath, DATA_FOLDER_NAME, SETTINGS_SUBFOLDER_NAME, PROGRAM_NAME )
    settings.load()

    application = BiblelatorSettingsEditor( tkRootWindow, homeFolderPath, loggingFolderPath, iconImage )
    # Calls to the window manager class (wm in Tk)
    #application.master.title( programNameVersion )
    #application.master.minsize( application.minimumXSize, application.minimumYSize )

    # Program a shutdown
    tkRootWindow.after( 30_000, tkRootWindow.destroy ) # Destroy the widget after 30 seconds

    # Start the program running
    tkRootWindow.mainloop()
# end of BiblelatorSettingsEditor.fullDemo

def main( homeFolderPath, loggingFolderPath ) -> None:
    """
    Main program to handle command line parameters and then run what they want.
    """
    BibleOrgSysGlobals.introduceProgram( __name__, programNameVersion, LAST_MODIFIED_DATE )

    #vPrint( 'Quiet', debuggingThisModule, 'FP main', repr(homeFolderPath), repr(loggingFolderPath) )

    numInstancesFound = 0
    if sys.platform == 'linux':
        myProcess = subprocess.Popen( ['ps','xa'], stdout=subprocess.PIPE, stderr=subprocess.PIPE )
        programOutputBytes, programErrorOutputBytes = myProcess.communicate()
        #vPrint( 'Quiet', debuggingThisModule, 'pob', programOutputBytes, programErrorOutputBytes )
        #returnCode = myProcess.returncode
        programOutputString = programOutputBytes.decode( encoding='utf-8', errors='replace' ) if programOutputBytes else None
        programErrorOutputString = programErrorOutputBytes.decode( encoding='utf-8', errors='replace' ) if programErrorOutputBytes else None
        #vPrint( 'Quiet', debuggingThisModule, 'processes', repr(programOutputString) )
        for line in programOutputString.split( '\n' ):
            if 'python' in line and PROGRAM_NAME+'.py' in line:
                if BibleOrgSysGlobals.debugFlag: vPrint( 'Quiet', debuggingThisModule, 'Found in ps xa:', repr(line) )
                numInstancesFound += 1
        if programErrorOutputString: logging.critical( "ps xa got error: {}".format( programErrorOutputString ) )
    elif sys.platform in ( 'win32', 'win64', ):
        myProcess = subprocess.Popen( ['tasklist.exe'], stdout=subprocess.PIPE, stderr=subprocess.PIPE )
        programOutputBytes, programErrorOutputBytes = myProcess.communicate()
        #vPrint( 'Quiet', debuggingThisModule, 'pob', programOutputBytes, programErrorOutputBytes )
        #returnCode = myProcess.returncode
        programOutputString = programOutputBytes.decode( encoding='utf-8', errors='replace' ) if programOutputBytes else None
        programErrorOutputString = programErrorOutputBytes.decode( encoding='utf-8', errors='replace' ) if programErrorOutputBytes else None
        #vPrint( 'Quiet', debuggingThisModule, 'processes', repr(programOutputString) )
        for line in programOutputString.split( '\n' ):
            if PROGRAM_NAME+'.py' in line:
                if BibleOrgSysGlobals.debugFlag: vPrint( 'Quiet', debuggingThisModule, 'Found in tasklist:', repr(line) )
                numInstancesFound += 1
        if programErrorOutputString: logging.critical( "tasklist got error: {}".format( programErrorOutputString ) )
    else: logging.critical( "Don't know how to check for already running instances in {}/{}.".format( sys.platform, os.name ) )
    if numInstancesFound > 1:
        import easygui
        logging.critical( "Found {} instances of {} running.".format( numInstancesFound, PROGRAM_NAME ) )
        result = easygui.ynbox('Seems {} might be already running: Continue?'.format( PROGRAM_NAME), programNameVersion, ('Yes', 'No'))
        if not result:
            logging.info( "Exiting as user requested." )
            sys.exit()

    tkRootWindow = tk.Tk()
    if BibleOrgSysGlobals.debugFlag:
        vPrint( 'Quiet', debuggingThisModule, 'Windowing system is', repr( tkRootWindow.tk.call('tk', 'windowingsystem') ) ) # e.g., 'x11'

    # Set the window icon and title
    iconImage = tk.PhotoImage( file=DATAFILES_FOLDERPATH.joinpath( 'Biblelator.gif' ) )
    tkRootWindow.tk.call( 'wm', 'iconphoto', tkRootWindow._w, iconImage )
    tkRootWindow.title( programNameVersion + ' ' + _('starting') + '…' )
    application = BiblelatorSettingsEditor( tkRootWindow, homeFolderPath, loggingFolderPath, iconImage )
    # Calls to the window manager class (wm in Tk)
    #application.master.title( programNameVersion )
    #application.master.minsize( application.minimumXSize, application.minimumYSize )

    # Start the program running
    tkRootWindow.mainloop()
# end of BiblelatorSettingsEditor.main

def run() -> None:
    """
    """
    multiprocessing.freeze_support() # Multiprocessing support for frozen Windows executables

    # Configure basic set-up
    homeFolderPath = BibleOrgSysGlobals.findHomeFolderPath()
    # if homeFolderPath[-1] not in '/\\': homeFolderPath += '/'
    # loggingFolderPath = os.path.join( homeFolderPath, DATA_FOLDER_NAME, LOGGING_SUBFOLDER_NAME )
    loggingFolderPath = homeFolderPath.joinpath( DATA_FOLDER_NAME, LOGGING_SUBFOLDER_NAME )
    parser = BibleOrgSysGlobals.setup( SHORT_PROGRAM_NAME, PROGRAM_VERSION, loggingFolderPath=loggingFolderPath )
    parser.add_argument( '-o', '--override', type=str, metavar='INIFilename', dest='override', help="override use of Biblelator.ini set-up" )
    BibleOrgSysGlobals.addStandardOptionsAndProcess( parser )
    #vPrint( 'Quiet', debuggingThisModule, BibleOrgSysGlobals.commandLineArguments ); halt

    if BibleOrgSysGlobals.debugFlag:
        vPrint( 'Quiet', debuggingThisModule, "Platform is", sys.platform ) # e.g., 'linux,'win32'
        vPrint( 'Quiet', debuggingThisModule, "OS name is", os.name ) # e.g., 'posix','nt'
        if sys.platform == "linux": vPrint( 'Quiet', debuggingThisModule, "OS uname is", os.uname() ) # gives about five fields
        vPrint( 'Quiet', debuggingThisModule, "Running main…" )

    main( homeFolderPath, loggingFolderPath )

    BibleOrgSysGlobals.closedown( PROGRAM_NAME, PROGRAM_VERSION )
# end of BiblelatorSettingsEditor.run

if __name__ == '__main__':
    run()
# end of BiblelatorSettingsEditor.py



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# AutocompleteFunctions.py
#
# Functions to support the autocomplete function in text editors
#
# Copyright (C) 2016-2018 Robert Hunt
# Author: Robert Hunt <Freely.Given.org+Biblelator@gmail.com>
# License: See gpl-3.0.txt
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Autocomplete is the ability to pop-up a little window of suggested word completions
    as the user is typing. The user can then accept a suggestion with ENTER, or
    close the window with ESC, or just to continue typing.

This module contains most of the helper functions for loading the autocomplete
    words (which may be from a Bible or from a dictionary, etc.)
"""

from gettext import gettext as _

LAST_MODIFIED_DATE = '2018-11-25' # by RJH
SHORT_PROGRAM_NAME = "AutocompleteFunctions"
PROGRAM_NAME = "Biblelator Autocomplete Functions"
PROGRAM_VERSION = '0.46'
programNameVersion = f'{PROGRAM_NAME} v{PROGRAM_VERSION}'

debuggingThisModule = False


import sys
import os
import logging
import multiprocessing
import time
from collections import defaultdict

import tkinter as tk

# Biblelator imports
if __name__ == '__main__':
    import sys
    aboveAboveFolderPath = os.path.dirname( os.path.dirname( os.path.dirname( os.path.abspath( __file__ ) ) ) )
    if aboveAboveFolderPath not in sys.path:
        sys.path.insert( 0, aboveAboveFolderPath )
from Biblelator.Windows.TextBoxes import TRAILING_SPACE_SUBSTITUTE, MULTIPLE_SPACE_SUBSTITUTE

# BibleOrgSys imports
from BibleOrgSys import BibleOrgSysGlobals
from BibleOrgSys.BibleOrgSysGlobals import vPrint
#from BibleOrgSys.Internals.InternalBibleInternals import BOS_PRINTABLE_MARKERS, BOS_EXTRA_TYPES
from BibleOrgSys.Reference.USFM3Markers import USFM_PRINTABLE_MARKERS



AVOID_BOOKS = ( 'FRT', 'BAK', 'GLS', 'XXA','XXB','XXC','XXD','XXE','XXF','XXG', 'NDX', 'UNK', )
END_CHARS_TO_REMOVE = ',—.–!?”:;' # NOTE: This intentionally doesn't include close parenthesis and similar
HUNSPELL_DICTIONARY_FOLDERS = ( '/usr/share/hunspell/', )



def setAutocompleteWords( editWindowObject, wordList, append=False ):
    """
    Given a word list, set the entries into the autocomplete words
        for an edit window and then do necessary house-keeping.

    Note that the original word order is preserved (if the supplied wordList has an order)
        so that more common/likely words can appear at the top of the list if desired.
    """
    logging.info( "AutocompleteFunctions.setAutocompleteWords( …, {}, {} )".format( len(wordList), append ) )
    if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
        #vPrint( 'Quiet', debuggingThisModule, "AutocompleteFunctions.setAutocompleteWords( {} )".format( wordList, append ) )
        vPrint( 'Quiet', debuggingThisModule, "AutocompleteFunctions.setAutocompleteWords( …, {}, {} )".format( len(wordList), append ) )
        editWindowObject.parentApp.setDebugText( "setAutocompleteWords…" )

    editWindowObject.parentApp.setWaitStatus( _("Setting autocomplete words…") )
    if not append: editWindowObject.autocompleteWords = {}

    for word in wordList:
        #if "'" not in word and '1' not in word:
            #if '(' in word and ')' not in word: # perhaps something like we(excl
                #word = word + ')' # append a matching/final parenthesis
        if len(word) >= editWindowObject.autocompleteMinLength:
            firstLetter, remainder = word[0], word[1:]
            if firstLetter not in editWindowObject.autocompleteWords: editWindowObject.autocompleteWords[firstLetter] = []
            if remainder in editWindowObject.autocompleteWords[firstLetter]:
                if 0 and BibleOrgSysGlobals.debugFlag and debuggingThisModule:
                    vPrint( 'Quiet', debuggingThisModule, "    setAutocompleteWords discarded {!r} duplicate".format( word ) )
            else: # not already in the list
                editWindowObject.autocompleteWords[firstLetter].append( remainder )
                for char in word:
                    if char not in editWindowObject.autocompleteWordChars:
                        if BibleOrgSysGlobals.debugFlag: assert char not in '\n\r'
                        if char not in ' .':
                            editWindowObject.autocompleteWordChars += char
                            if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
                                vPrint( 'Quiet', debuggingThisModule, "    setAutocompleteWords added {!r} as new wordChar".format( char ) )
        #elif BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            #vPrint( 'Quiet', debuggingThisModule, "    setAutocompleteWords discarded {!r} as too short".format( word ) )
        #elif BibleOrgSysGlobals.debugFlag and debuggingThisModule:
            #if "'" not in word:
                #vPrint( 'Quiet', debuggingThisModule, "    setAutocompleteWords discarded {!r} as unwanted".format( word ) )

    if BibleOrgSysGlobals.debugFlag and debuggingThisModule: # write wordlist
        vPrint( 'Quiet', debuggingThisModule, "  setAutocompleteWords: Writing autocomplete words to file…" )
        sortedKeys = sorted( editWindowObject.autocompleteWords.keys() )
        with open( 'autocompleteWordList.txt', 'wt', encoding='utf-8' ) as wordFile:
            wordCount = 0
            for firstLetter in sortedKeys:
                for remainder in sorted( editWindowObject.autocompleteWords[firstLetter] ):
                    wordFile.write( firstLetter+remainder )
                    wordCount += 1
                    if wordCount == 8: wordFile.write( '\n' ); wordCount = 0
                    else: wordFile.write( ' ' )

    if BibleOrgSysGlobals.debugFlag: # print detailed stats
        sortedKeys = sorted( editWindowObject.autocompleteWords.keys() )
        vPrint( 'Never', debuggingThisModule, "  autocomplete first letters", len(editWindowObject.autocompleteWords), sortedKeys )
        grandtotal = 0
        wordNumTotals = defaultdict( int )
        for firstLetter in sortedKeys:
            total = len(editWindowObject.autocompleteWords[firstLetter])
            for wordRemainder in editWindowObject.autocompleteWords[firstLetter]:
                wordNumTotals[wordRemainder.count(' ')] += 1
            if debuggingThisModule:
                vPrint( 'Quiet', debuggingThisModule, "    {!r} {:,}{}" \
                    .format( firstLetter, total, '' if total>19 else ' '+str(editWindowObject.autocompleteWords[firstLetter]) ) )
            grandtotal += total
        #if BibleOrgSysGlobals.debugFlag or BibleOrgSysGlobals.verbosityLevel > 1:
        vPrint( 'Quiet', debuggingThisModule, "  autocomplete total words loaded = {:,}".format( grandtotal ) )
        if debuggingThisModule:
            for spaceCount in wordNumTotals:
                vPrint( 'Quiet', debuggingThisModule, "    {} words: {}".format( spaceCount+1, wordNumTotals[spaceCount] ) )

    editWindowObject.parentApp.setReadyStatus()
# end of AutocompleteFunctions.setAutocompleteWords



internalMarkers = None
DUMMY_VALUE = 999999 # Some number bigger than the number of characters in a line

def countBookWords( BBB, internalBible, filename, isCurrentBook, internalMarkers ):
    """
    Find all the words in the Bible book and their usage counts.

    Note that this function doesn't use the internalBible books
        but rather loads the USFM (text) files directly.

    Note also that the internalMarkersList has to be passed as a paramter,
        because multi-processing on Windows can't access global variables.

    Returns a dictionary containing the results for the book.
    """
    logging.debug( "countBookWords( {}, {}, {} )".format( BBB, internalBible, filename ) )
    if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
        vPrint( 'Quiet', debuggingThisModule, "countBookWords( {}, {}, {} )".format( BBB, internalBible, filename ) )
    if BBB in AVOID_BOOKS:
        #vPrint( 'Quiet', debuggingThisModule, "Didn't load autocomplete words from {} {}".format( internalBible.getAName(), BBB ) )
        return # Sometimes these books contain words from other languages, etc.

    countIncrement = 3 if isCurrentBook else 1 # Each word in current book counts higher so appears higher in the list
    # NOTE: This idea fails as soon as they change books in the edit window
    #       as the word lists are only loaded once at startup. (A reasonable compromise I think.)

    encoding = None
    if encoding is None: encoding = 'utf-8'
    lastLine, lineCount, lineDuples, lastMarker = '', 0, [], None
    wordCounts = defaultdict( int )

    def countWords( textLine ):
        """
        Note: Punctuation etc. is NOT removed.
        """
        #vPrint( 'Quiet', debuggingThisModule, "countWords( {!r} )".format( textLine ) )
        if '\\' in textLine: # we have internal markers to remove
            #vPrint( 'Quiet', debuggingThisModule, "  INT", marker, textLine )
            for iMarker in internalMarkers:
                #vPrint( 'Quiet', debuggingThisModule, "   GOT", repr(iMarker) )
                textLine = textLine.replace( iMarker+' ',' ' ).replace( iMarker+'*',' ' )
                if not '\\' in textLine: break
            #vPrint( 'Quiet', debuggingThisModule, "  NOW", marker, textLine )
        words = textLine.replace('—','— ').replace('–','– ').split() # Treat em-dash and en-dash as word break characters

        # Now look for (and count) single and some multiple word sequences
        for wx,word in enumerate( words ):
            if not word: continue
            if 'XXX' in word: continue # This is used in the Matigsalug project to mark errors
            singleWord = word
            while singleWord and singleWord[-1] in END_CHARS_TO_REMOVE:
                singleWord = singleWord[:-1] # Remove certain final punctuation
            if len(singleWord) > 2: wordCounts[singleWord] += countIncrement

            #if word[-1] not in '—.–':
            if wx < len(words)-1: # it's not the last word in the line
                doubleWord = word+' '+words[wx+1]
                #vPrint( 'Quiet', debuggingThisModule, 'doubleWord', repr(doubleWord) )
                adjustedDoubleWord = doubleWord[:-1] if doubleWord[-1] in END_CHARS_TO_REMOVE else doubleWord
                if '. ' not in adjustedDoubleWord: # don't go across sentence boundaries
                    wordCounts[adjustedDoubleWord] += countIncrement
                if wx < len(words)-2: # there's still two words after this one
                    tripleWord = doubleWord+' '+words[wx+2]
                    #vPrint( 'Quiet', debuggingThisModule, 'tripleWord', repr(tripleWord) )
                    adjustedTripleWord = tripleWord[:-1] if tripleWord[-1] in END_CHARS_TO_REMOVE else tripleWord
                    if '. ' not in adjustedTripleWord: # don't go across sentence boundaries
                        wordCounts[adjustedTripleWord] += countIncrement
                    if wx < len(words)-3: # there's still three words after this one
                        quadWord = tripleWord+' '+words[wx+3]
                        #vPrint( 'Quiet', debuggingThisModule, 'quadWord', repr(quadWord) )
                        adjustedQuadWord = quadWord[:-1] if quadWord[-1] in END_CHARS_TO_REMOVE else quadWord
                        if '. ' not in adjustedQuadWord: # don't go across sentence boundaries
                            wordCounts[adjustedQuadWord] += countIncrement
                        if wx < len(words)-4: # there's still four words after this one
                            quinWord = quadWord+' '+words[wx+4]
                            #vPrint( 'Quiet', debuggingThisModule, 'quinWord', repr(quinWord) )
                            adjustedQuinWord = quinWord[:-1] if quinWord[-1] in END_CHARS_TO_REMOVE else quinWord
                            if '. ' not in adjustedQuinWord: # don't go across sentence boundaries
                                wordCounts[adjustedQuinWord] += countIncrement
    # end of countWords

    # main code for countBookWords
    USFMFilepath = os.path.join( internalBible.sourceFolder, filename )
    with open( USFMFilepath, 'rt', encoding=internalBible.encoding ) as bookFile:
        try:
            for line in bookFile:
                lineCount += 1
                if lineCount==1 and encoding.lower()=='utf-8' and line[0]==chr(65279): #U+FEFF
                    logging.info( "countBookWords: Detected Unicode Byte Order Marker (BOM) in {}".format( USFMFilepath ) )
                    line = line[1:] # Remove the Unicode Byte Order Marker (BOM)
                if line and line[-1]=='\n': line=line[:-1] # Removing trailing newline character
                if not line: continue # Just discard blank lines
                lastLine = line
                #vPrint( 'Quiet', debuggingThisModule, 'USFM file line is {!r}'.format( line ) )
                #if line[0:2]=='\\_': continue # Just discard Toolbox header lines
                if line[0]=='#': continue # Just discard comment lines

                if line[0]!='\\': # Not a SFM line
                    if lastMarker is None: # We don't have any SFM data lines yet
                        logging.error( "countBookWords: Non-USFM line in {} -- line ignored at #{}".format( USFMFilepath, lineCount) )
                        #vPrint( 'Quiet', debuggingThisModule, "SFMFile.py: XXZXResult is", lineDuples, len(line) )
                        #for x in range(0, min(6,len(line))):
                            #vPrint( 'Quiet', debuggingThisModule, x, "'" + str(ord(line[x])) + "'" )
                        #raise IOError('Oops: Line break on last line ??? not handled here "' + line + '"')
                    else: # Append this continuation line
                        if lastMarker in USFM_PRINTABLE_MARKERS:
                            #oldmarker, oldtext = lineDuples.pop()
                            #vPrint( 'Quiet', debuggingThisModule, "Popped",oldmarker,oldtext)
                            #vPrint( 'Quiet', debuggingThisModule, "Adding", line, "to", oldmarker, oldtext)
                            #lineDuples.append( (oldmarker, oldtext+' '+line) )
                            countWords( line )
                        continue

                lineAfterBackslash = line[1:]
                si1 = lineAfterBackslash.find( ' ' )
                si2 = lineAfterBackslash.find( '*' )
                si3 = lineAfterBackslash.find( '\\' )
                if si1==-1: si1 = DUMMY_VALUE
                if si2==-1: si2 = DUMMY_VALUE
                if si3==-1: si3 = DUMMY_VALUE
                si = min( si1, si2, si3 )

                if si != DUMMY_VALUE:
                    if si == si3: # Marker stops before a backslash
                        marker = lineAfterBackslash[:si3]
                        text = lineAfterBackslash[si3:]
                    elif si == si2: # Marker stops at an asterisk
                        marker = lineAfterBackslash[:si2+1]
                        text = lineAfterBackslash[si2+1:]
                    elif si == si1: # Marker stops before a space
                        marker = lineAfterBackslash[:si1]
                        text = lineAfterBackslash[si1+1:] # We drop the space completely
                else: # The line is only the marker
                    marker = lineAfterBackslash
                    text = ''

                #vPrint( 'Quiet', debuggingThisModule, " ", repr(marker), repr(text) )
                #if marker not in ignoreSFMs:
                if marker in USFM_PRINTABLE_MARKERS and text:
                    #vPrint( 'Quiet', debuggingThisModule, "   1", marker, text )
                    if marker == 'v' and text[0].isdigit():
                        try: text = text.split( None, 1 )[1]
                        except IndexError: text = ''
                    #vPrint( 'Quiet', debuggingThisModule, "   2", marker, text )
                    countWords( text )
                    #if not lineDuples: # Just for detection of start of real USFM
                        #lineDuples.append( (marker, text) )
                lastMarker = marker

        except UnicodeError as err:
            vPrint( 'Quiet', debuggingThisModule, "Unicode error:", sys.exc_info()[0], err )
            logging.critical( "countBookWords: Invalid line in {} -- line ignored at #{}".format( USFMFilepath, lineCount) )
            if lineCount > 1: vPrint( 'Quiet', debuggingThisModule, 'Previous line was: ', lastLine )
            #vPrint( 'Quiet', debuggingThisModule, line )
            #raise

    return wordCounts
# end of AutocompleteFunctions.countBookWords


def countBookWordsHelper( parameters ):
    """
    Parameter parameters is a 4-tuple containing the BBB, folder, filename, currentBook flag, and internalMarkersList

    The internalMarkers have to be passed, because multi-processing on Windows can't access global variables.
    """
    return countBookWords( *parameters )
# end of AutocompleteFunctions.countBookWordsHelper


def loadBibleBookAutocompleteWords( editWindowObject ):
    """
    Load all the existing words in a USFM or Paratext Bible book
        to fill the autocomplete mechanism

    editWindowObject here is a USFM or ESFM edit window.

    NOTE: This list should theoretically be updated as the user enters new words!
    """
    logging.info( "loadBibleBookAutocompleteWords()" )
    if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
        vPrint( 'Quiet', debuggingThisModule, "loadBibleBookAutocompleteWords()" )
        editWindowObject.parentApp.setDebugText( "loadBibleBookAutocompleteWords…" )

    editWindowObject.parentApp.setWaitStatus( _("Loading {} Bible book words…").format( editWindowObject.projectName ) )
    currentBBB = editWindowObject.currentVerseKey.getBBB()
    #vPrint( 'Quiet', debuggingThisModule, "  got BBB", repr(BBB) )
    if currentBBB == 'UNK': return # UNKnown book -- no use here

    if not editWindowObject.internalBible.preloadDone: editWindowObject.internalBible.preload()
    foundFilename = None
    for BBB2,filename in editWindowObject.internalBible.maximumPossibleFilenameTuples:
        if BBB2 == currentBBB: foundFilename = filename; break

    wordCountResults = countBookWords( currentBBB, editWindowObject.internalBible, foundFilename, False )
    #vPrint( 'Quiet', debuggingThisModule, 'wordCountResults', len(wordCountResults) )

    # Would be nice to load current book first, but we don't know it yet
    autocompleteWords = []
    #if BibleOrgSysGlobals.debugFlag:
        #autocompleteWords = [ 'Lord God', 'Lord your(pl) God', '(is)', '(are)', '(were)', '(one who)', ]
    try:
        # Sort the word-list for the book to put the most common words first
        #vPrint( 'Quiet', debuggingThisModule, 'wordCountResults', currentBBB, discoveryResults )
        #vPrint( 'Quiet', debuggingThisModule, currentBBB, 'mTWC', discoveryResults['mainTextWordCounts'] )
        #qqq = sorted( discoveryResults['mainTextWordCounts'].items(), key=lambda c: -c[1] )
        #vPrint( 'Quiet', debuggingThisModule, 'qqq', qqq )
        for word,count in sorted( wordCountResults.items(),
                                key=lambda duple: -duple[1] ):
            if len(word) >= editWindowObject.autocompleteMinLength \
            and word not in autocompleteWords: # just in case we had some (common) words in there already
                if ' ' not in word or count > 4:
                    autocompleteWords.append( word )
                #else: vPrint( 'Quiet', debuggingThisModule, 'loadBibleBookAutocompleteWords discarding', repr(word) )
    except KeyError:
        vPrint( 'Quiet', debuggingThisModule, "Why did {} have no words???".format( currentBBB ) )
        #pass # Nothing for this book
    #vPrint( 'Quiet', debuggingThisModule, 'autocompleteWords', len(autocompleteWords) )
    setAutocompleteWords( editWindowObject, autocompleteWords )
    editWindowObject.addAllNewWords = True
# end of AutocompleteFunctions.loadBibleBookAutocompleteWords



def loadBibleAutocompleteWords( editWindowObject ):
    """
    Load all the existing words in a USFM or Paratext Bible Project
        to fill the autocomplete mechanism.

    This is rather slow because of course, the entire Bible has to be read and processed first.

    editWindowObject here is a USFM or ESFM edit window.

    NOTE: This list should theoretically be updated as the user enters new words!
    """
    startTime = time.time()
    if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
        vPrint( 'Quiet', debuggingThisModule, "AutocompleteFunctions.loadBibleAutocompleteWords()" )
        editWindowObject.parentApp.setDebugText( "loadBibleAutocompleteWords…" )

    global internalMarkers
    if internalMarkers is None: # Get our list of markers -- note that the more common note markers are first
        internalMarkers = BibleOrgSysGlobals.loadedUSFMMarkers.getNoteMarkersList() \
            + BibleOrgSysGlobals.loadedUSFMMarkers.getCharacterMarkersList( includeBackslash=False, includeEndMarkers=False, includeNestedMarkers=True, expandNumberableMarkers=True )
        internalMarkers = ['\\'+marker for marker in internalMarkers]

    editWindowObject.parentApp.setWaitStatus( _("Loading {} Bible words…").format( editWindowObject.projectName ) )
    currentBBB = editWindowObject.currentVerseKey.getBBB()
    if BibleOrgSysGlobals.debugFlag and debuggingThisModule: vPrint( 'Quiet', debuggingThisModule, "  got current BBB", repr(currentBBB) )

    if not editWindowObject.internalBible.preloadDone: editWindowObject.internalBible.preload()
    bookWordCounts = {}
    if editWindowObject.internalBible.maximumPossibleFilenameTuples:
        if BibleOrgSysGlobals.maxProcesses > 1: # Load all the books as quickly as possible
            parameters = [(BBB,editWindowObject.internalBible,filename,BBB==currentBBB,internalMarkers) for BBB,filename in editWindowObject.internalBible.maximumPossibleFilenameTuples] # Can only pass a single parameter to map
            vPrint( 'Normal', debuggingThisModule, "Autocomplete: loading up to {} USFM books using {} processes…".format( len(editWindowObject.internalBible.maximumPossibleFilenameTuples), BibleOrgSysGlobals.maxProcesses ) )
            vPrint( 'Normal', debuggingThisModule, "  NOTE: Outputs (including error & warning messages) from loading words from BibleOrgSys.Bible books may be interspersed." )
            BibleOrgSysGlobals.alreadyMultiprocessing = True
            with multiprocessing.Pool( processes=BibleOrgSysGlobals.maxProcesses ) as pool: # start worker processes
                results = pool.map( countBookWordsHelper, parameters ) # have the pool do our loads
                assert len(results) == len(editWindowObject.internalBible.maximumPossibleFilenameTuples)
                for (BBB,filename),counts in zip( editWindowObject.internalBible.maximumPossibleFilenameTuples, results ):
                    #vPrint( 'Quiet', debuggingThisModule, "XX", BBB, filename, len(counts) if counts else counts )
                    bookWordCounts[BBB] = counts
                BibleOrgSysGlobals.alreadyMultiprocessing = False
        else: # Just single threaded
            # Load the books one by one -- assuming that they have regular Paratext style filenames
            for BBB,filename in editWindowObject.internalBible.maximumPossibleFilenameTuples:
                #if BibleOrgSysGlobals.verbosityLevel>1 or BibleOrgSysGlobals.debugFlag:
                    #vPrint( 'Quiet', debuggingThisModule, _("  USFMBible: Loading {} from {} from {}…").format( BBB, editWindowObject.internalBible.getAName(), editWindowObject.internalBible.sourceFolder ) )
                bookWordCounts[BBB] = countBookWords( BBB, editWindowObject.internalBible, filename, BBB==currentBBB ) # also saves it
    else:
        logging.critical( "Autocomplete: " + _("No books to load in folder '{}'!").format( editWindowObject.internalBible.sourceFolder ) )

    # Now combine the books
    autocompleteCounts = {}
    for BBB,counts in bookWordCounts.items(): # combine word counts for all books
        #vPrint( 'Quiet', debuggingThisModule, "here", BBB, len(counts) )
        if counts:
            for word, count in counts.items():
                #vPrint( 'Quiet', debuggingThisModule, "  ", word, count )
                if len(word) >= editWindowObject.autocompleteMinLength:
                    if word in autocompleteCounts: autocompleteCounts[word] += count
                    else: autocompleteCounts[word] = count
    #vPrint( 'Quiet', debuggingThisModule, "there", len(autocompleteCounts) )

    # Now make our list sorted with most common words first
    autocompleteWords = []
    #if BibleOrgSysGlobals.debugFlag: # add some multi-word entries just for testing
        #autocompleteWords = [ 'Lord God', 'Lord your(pl) God', '(is)', '(are)', '(were)', '(one who)', ]
    for word,count in sorted( autocompleteCounts.items(), key=lambda duple: -duple[1] ):
        if ' ' not in word or count > 9:
            autocompleteWords.append( word )
        #else:
            #vPrint( 'Quiet', debuggingThisModule, 'loadBibleAutocompleteWords discarding', repr(word) )
            #if ' ' not in word: halt
    #if BibleOrgSysGlobals.debugFlag and debuggingThisModule: vPrint( 'Quiet', debuggingThisModule, 'acW', autocompleteWords )

    #vPrint( 'Quiet', debuggingThisModule, 'autocompleteWords', len(autocompleteWords) )
    setAutocompleteWords( editWindowObject, autocompleteWords )
    if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
        vPrint( 'Quiet', debuggingThisModule, "loadBibleAutocompleteWords took", time.time()-startTime )
    editWindowObject.addAllNewWords = True
# end of AutocompleteFunctions.loadBibleAutocompleteWords



def loadHunspellAutocompleteWords( editWindowObject, dictionaryFilepath, encoding='utf-8' ):
    """
    Load all the existing words in a Hunspell-type dictionary
        to fill the autocomplete mechanism

    editWindowObject here is a text edit window or derivation.

    NOTE: This list maybe should be updated as the user enters new words
        or else have an additional user dictionary.
    """
    logging.info( "loadHunspellAutocompleteWords( {}, {} )".format( dictionaryFilepath, encoding ) )
    if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
        vPrint( 'Quiet', debuggingThisModule, "loadHunspellAutocompleteWords( {}, {} )".format( dictionaryFilepath, encoding ) )
        editWindowObject.parentApp.setDebugText( "loadHunspellAutocompleteWords…" )

    editWindowObject.parentApp.setWaitStatus( _("Loading dictionary…") )
    internalCount = None
    autocompleteWords = []
    lineCount = 0
    with open( dictionaryFilepath, 'rt', encoding=encoding ) as dictionaryFile:
        for line in dictionaryFile:
            lineCount += 1
            if lineCount==1 and encoding.lower()=='utf-8' and line[0]==chr(65279): #U+FEFF or \ufeff
                logging.info( "loadHunspellAutocompleteWords: Detected Unicode Byte Order Marker (BOM) in {}".format( dictionaryFilepath ) )
                line = line[1:] # Remove the Unicode Byte Order Marker (BOM)
            if line and line[-1]=='\n': line=line[:-1] # Remove trailing newline character
            if not line: continue # Just discard blank lines
            #vPrint( 'Quiet', debuggingThisModule, "line", lineCount, repr(line) )

            if lineCount==1 and line.isdigit(): # first line seems to be a count
                internalCount = int( line )
                continue

            try: word, codes = line.split( '/', 1 )
            except ValueError: word, codes = line, ''
            if word in ('3GPP','AA','ACAS',): continue # Throw out rubbish
            #vPrint( 'Quiet', debuggingThisModule, "word", repr(word), repr(codes) )
            autocompleteWords.append( word )

            wordDeleteA = word[:-1] if word[-1]=='a' else word
            wordDeleteE = word[:-1] if word[-1]=='e' else word
            wordDeleteEY = word[:-1] if word[-1] in ('e','y',) else word
            wordDeleteEChangeY = wordDeleteE[:-1]+'i' if wordDeleteE[-1]=='y' else wordDeleteE
            wordAddEAfterSY = word+'e' if word[-1]=='s' else word
            wordAddEAfterSY = wordAddEAfterSY[:-1]+'ie' if wordAddEAfterSY[-1]=='y' else wordAddEAfterSY
            wordYtoI = word[:-1]+'i' if word[-1]=='y' else word

            generatedWords = []
            """
            B -able, -ability, last syllable of stem stressed, -ate words &gt; 2 syllables
            b -ible, very basic rules, only dropped e
            D -ed, regular verb past tenses, last syllable of stem stressed
            d -ed, -ing, regular verb past tenses and adverbial form, last syllable NOT stressed
            E dis- Prefix for negation
            e out- Prefix
            h -edly, adverbial, simplified rules
            I in- im- il- ir- Prefix, opposite of.
            i -edness, degree, simplified rules
            j -fully, suffix
            K pre-, prefix
            k -ingly, adverbial form, simplified rules
            L -ment, -ments, -ment's, suffix, both generated
            l -ably, simplified rules
            N -ion, noun from verb, stress on last syllable of stem
            n -ion, -ions, noun from verb, stress NOT on last syllable of stem
            q -isation, -isations, -ization, -izations, all generated
            S -s, noun plurals, verb conjugation
            s -iser, -isers, -izer, -izers, -iser's, -izer's, all generated
            T -er, -est, adjectival comparatives, both generated
            t -isable, -isability, -izable, -izability, all generated
            u -iveness, ending for verbs
            V -ive, ending for verbs (simplified rules)
            v -ively, ending for verbs
            W -ic, adjectival ending, simplified rules
            w -ical, adjectival ending, simplified rules
            X -ions, noun plural, stress on last syllable of stem, simplified rules
            x -ional, -ionally, simplified rules, both endings formed
            Y -ly, adverb endings for adjectives
            y -ry, adjectival and noun forms, simplified rules.
            0 -al, noun from verb, simplified rules
            1 -ically, adverbial double suffix, simplified rules
            2 -iness, y+ness ending, simplified rules
            3 -ist, -ists, -ists's, professions
            5 -woman, -women, -woman's suffixes, all generated
            7 -able, last syllable NOT stressed, -ate words <= 2 syllables
            """
            for code in codes:
                #vPrint( 'Quiet', debuggingThisModule, "  code", code, "for", repr(word) )
                if code == 'A': generatedWords.append( 're' + word )
                elif code == 'a': generatedWords.append( 'mis' + word )
                elif code == 'B': generatedWords.append( word+'able' ); generatedWords.append( word+'ability' )
                elif code == 'b': generatedWords.append( wordDeleteE + 'ible' ); generatedWords.append( wordDeleteE + 'ibility' )
                elif code == 'C': generatedWords.append( 'de' + word )
                elif code == 'c': generatedWords.append( 'over' + word )
                elif code == 'D': generatedWords.append( wordDeleteE + 'ed' ) # last syllable of stem stressed
                elif code == 'd': generatedWords.append( word+'ed' ); generatedWords.append( word+'ing' )
                elif code == 'E': generatedWords.append( 'dis' + word )
                elif code == 'e': generatedWords.append( 'out' + word )
                elif code == 'F': # e.g., ment -> prefix
                    if word[0] in ( 'm','b','p',): generatedWords.append( 'com' + word )
                    #elif word[0] == 'l': generatedWords.append( 'il' + word )
                    #elif word[0] == 'r': generatedWords.append( 'ir' + word )
                    else: generatedWords.append( 'con' + word )
                elif code == 'f': generatedWords.append( 'under' + word )
                elif code == 'G': # e.g., XXX -> ending for verbs, stress on last syllable of stem
                    generatedWords.append( wordDeleteE + 'ing' )
                elif code == 'g': # e.g., palate -> last syllable NOT stressed
                    generatedWords.append( wordDeleteE + 'ability' )
                elif code == 'H': # e.g., eighty-four -> number specific suffixes, both generated
                    generatedWords.append( word + 'th' ); generatedWords.append( word + 'fold' )
                elif code == 'h': # e.g., abash -> adverbial, simplified rules
                    generatedWords.append( wordDeleteE + 'edly' )
                elif code == 'I':
                    if word[0] in ( 'm','b','p',): generatedWords.append( 'im' + word )
                    elif word[0] == 'l': generatedWords.append( 'il' + word )
                    elif word[0] == 'r': generatedWords.append( 'ir' + word )
                    else: generatedWords.append( 'in' + word )
                elif code == 'i': generatedWords.append( wordDeleteEY + 'edness' )
                elif code == 'J': # e.g., band -> plural noun version of verb ing ending, simplified rules
                    generatedWords.append( word + 'ings' )
                elif code == 'j': # e.g., bliss, wonder -> suffix
                    generatedWords.append( word + 'fully' )
                elif code == 'K': generatedWords.append( 'pre' + word )
                elif code == 'k': generatedWords.append( wordDeleteE + 'ingly' )
                elif code == 'L': generatedWords.append( word+'ment' ); generatedWords.append( word+'ments' ); generatedWords.append( word+"ment's" )
                elif code == 'l': # e.g., avoid
                    generatedWords.append( word + 'ably' )
                elif code == 'M': # e.g., abalone -> possessive form
                    generatedWords.append( word + "'s" ) # What about other apostrophe types
                elif code == 'm': # e.g., artillery -> suffixes, all generated
                    generatedWords.append( word+'man' ); generatedWords.append( word+"man's" ); generatedWords.append( word+'men' ); generatedWords.append( word+"men's" )
                elif code == 'N': # e.g., assume
                    generatedWords.append( wordDeleteE + 'ion' )
                elif code == 'n': generatedWords.append( wordDeleteE+'ion' ); generatedWords.append( wordDeleteE+'ions' )
                elif code == 'O': # e.g., fiction -> prefix
                    generatedWords.append( 'non' + word )
                elif code == 'o': # e.g., apocrypha -> adverb from verb, simplified rules
                    generatedWords.append( wordDeleteA + 'ally' )
                elif code == 'P': # e.g., absolute -> adjective degree of comparison
                    generatedWords.append( wordYtoI+'ness' ); generatedWords.append( wordYtoI+"ness's" )
                elif code == 'p': # e.g., body -> comparative suffix
                    generatedWords.append( word+'less' )
                elif code == 'Q': # e.g., pre -> all generated
                    generatedWords.append( wordDeleteE+'ise' ); generatedWords.append( wordDeleteE+'ised' ); generatedWords.append( wordDeleteE+'ises' ); generatedWords.append( wordDeleteE+'ising' )
                    generatedWords.append( wordDeleteE+'ize' ); generatedWords.append( wordDeleteE+'ized' ); generatedWords.append( wordDeleteE+'izes' ); generatedWords.append( wordDeleteE+'izing' )
                elif code == 'R': # e.g., abjure -> doer, last syllable stressed, both forms generated
                    generatedWords.append( wordDeleteE+'er' ); generatedWords.append( wordDeleteE+'ers' ); generatedWords.append( wordDeleteE+"er's" )
                elif code == 'r': # e.g., backslid -> doer, last syllable NOT stressed, both forms generated
                    generatedWords.append( word+'er' ); generatedWords.append( word+'ers' ); generatedWords.append( word+"er's" )
                elif code == 'S': generatedWords.append( wordAddEAfterSY + 's' )
                elif code == 'T': generatedWords.append( wordDeleteEY+'er' ); generatedWords.append( wordDeleteEY+'est' )
                elif code == 'U': generatedWords.append( 'un' + word )
                elif code == 'u': generatedWords.append( wordDeleteEY + 'iveness' )
                elif code == 'V': generatedWords.append( wordDeleteEY + 'ive' )
                elif code == 'v': generatedWords.append( wordDeleteEY + 'ively' )
                elif code == 'W': generatedWords.append( wordDeleteEY + 'ic' )
                elif code == 'w': generatedWords.append( wordDeleteEY + 'ical' )
                elif code == 'X': # e.g., assume
                    generatedWords.append( wordDeleteEY + 'ions' )
                elif code == 'x': generatedWords.append( wordDeleteEY + 'ional' ); generatedWords.append( word + 'ionally' )
                elif code == 'Y': generatedWords.append( wordYtoI + 'ly' )
                elif code == 'y': generatedWords.append( word + 'ry' )
                elif code == 'Z': # e.g., academe -> diminutive and adjectival form, simplified rules
                    generatedWords.append( wordDeleteE + 'y' )
                elif code == 'z': # e.g., cage -> adverbial ending where adjective adds y
                    generatedWords.append( wordDeleteEY + 'ily' )
                elif code == '1': vPrint( 'Quiet', debuggingThisModule, ' 1 on', repr(word), 'ignored' )
                elif code == '2': # e.g., bone -> y+ness ending, simplified rules
                    generatedWords.append( wordDeleteEY + 'iness' )
                elif code == '3': generatedWords.append( wordDeleteEY+'ist' ); generatedWords.append( wordDeleteE+'ists' ); generatedWords.append( wordDeleteE+"ist's" )
                elif code == '5': # e.g., chair
                    generatedWords.append( word+'woman' ); generatedWords.append( word+"woman's" ); generatedWords.append( word+'women' ); generatedWords.append( word+"women's" )
                elif code == '4': generatedWords.append( 'trans' + word )
                elif code == '6': # e.g., bliss, wonder -> suffix
                    generatedWords.append( word + 'ful' )
                elif code == '7': generatedWords.append( word + 'able' )
                elif BibleOrgSysGlobals.debugFlag:
                    vPrint( 'Quiet', debuggingThisModule, lineCount, "code", code, "for", repr(word), repr(codes) )
                    halt
            vPrint( 'Quiet', debuggingThisModule, "  generated", generatedWords, 'from', repr(word), repr(codes) )
            autocompleteWords.extend( generatedWords )

            #lastLine = line
            #if lineCount > 60: break
    #vPrint( 'Quiet', debuggingThisModule, 'acW', len(autocompleteWords), autocompleteWords )

    if editWindowObject.autocompleteMinLength < 4:
        vPrint( 'Quiet', debuggingThisModule, "NOTE: Lengthened autocompleteMinLength from {} to {}".format( editWindowObject.autocompleteMinLength, 4 ) )
        editWindowObject.autocompleteMinLength = 4 # Show the window after this many characters have been typed
    setAutocompleteWords( editWindowObject, autocompleteWords )
    editWindowObject.addAllNewWords = False
# end of AutocompleteFunctions.loadHunspellAutocompleteWords



def loadILEXAutocompleteWords( editWindowObject, dictionaryFilepath, lgCodes=None ):
    """
    Load all the existing words in an ILEX dictionary
        to fill the autocomplete mechanism

    editWindowObject here is a text edit window or derivation.

    NOTE: This list maybe should be updated as the user enters new words
        or else have an additional user dictionary.
    """
    logging.info( "loadILEXAutocompleteWords( {}, {} )".format( dictionaryFilepath, lgCodes ) )
    if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
        vPrint( 'Quiet', debuggingThisModule, "loadILEXAutocompleteWords( {}, {} )".format( dictionaryFilepath, lgCodes ) )
        editWindowObject.parentApp.setDebugText( "loadILEXAutocompleteWords…" )

    editWindowObject.parentApp.setWaitStatus( _("Loading dictionary…") )
    autocompleteWords = []
    lineCount = 0
    with open( dictionaryFilepath, 'rt', encoding='utf-8' ) as dictionaryFile:
        for line in dictionaryFile:
            lineCount += 1
            if lineCount==1:
                if line[0]==chr(65279): #U+FEFF
                    logging.info( "loadILEXAutocompleteWords1: Detected Unicode Byte Order Marker (BOM) in {}".format( dictionaryFilepath ) )
                    line = line[1:] # Remove the UTF-16 Unicode Byte Order Marker (BOM)
                elif line[:3] == 'ï»¿': # 0xEF,0xBB,0xBF
                    logging.info( "loadILEXAutocompleteWords2: Detected Unicode Byte Order Marker (BOM) in {}".format( dictionaryFilepath ) )
                    line = line[3:] # Remove the UTF-8 Unicode Byte Order Marker (BOM)
            if line and line[-1]=='\n': line=line[:-1] # Remove trailing newline character
            if not line: continue # Just discard blank lines
            #vPrint( 'Quiet', debuggingThisModule, "line", lineCount, repr(line) )

            # wd, lg, ps, and sc are the four compulsory fields in each record
            if line.startswith( '\\wd ' ):
                word = line[4:]
                if '*' in word and word[-2] == '*' and word[-1].isdigit(): # It has a subscript
                    word = word[:-2]
            elif line.startswith( '\\lg ' ):
                lgCode = line[4:]
                assert len(lgCode) == 3
            elif line.startswith( '\\ps ' ):
                POS = line[4:]

                if lgCodes is None or lgCode in lgCodes:
                    if POS != 'x': # abbreviations like AFAIK
                        if word not in autocompleteWords:
                            autocompleteWords.append( word )

            #lastLine = line
            #if lineCount > 600: break
    #vPrint( 'Quiet', debuggingThisModule, 'acW', len(autocompleteWords), autocompleteWords )

    if editWindowObject.autocompleteMinLength < 4:
        vPrint( 'Quiet', debuggingThisModule, "NOTE: Lengthened autocompleteMinLength from {} to {}".format( editWindowObject.autocompleteMinLength, 4 ) )
        editWindowObject.autocompleteMinLength = 4 # Show the window after this many characters have been typed
    setAutocompleteWords( editWindowObject, autocompleteWords )
    editWindowObject.addAllNewWords = False
# end of AutocompleteFunctions.loadILEXAutocompleteWords



############################################################################
#
# The following functions are part of the autocomplete code that's
#   executed while typing.
#
# self in these functions is assumed to be an edit window
#   containing self.textBox
#
############################################################################


def getCharactersBeforeCursor( self, charCount=1 ):
    """
    Needed for auto-correct functions.
    """
    #if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
        #vPrint( 'Quiet', debuggingThisModule, "AutocompleteFunctions.getCharactersBeforeCursor( {} )".format( charCount ) )

    previousText = self.textBox.get( tk.INSERT+'-{}c'.format( charCount ), tk.INSERT )
    #vPrint( 'Quiet', debuggingThisModule, 'getCharactersBeforeCursor: returning previousText', repr(previousText) )
    return previousText
# end of AutocompleteFunctions.getCharactersBeforeCursor


def getWordCharactersBeforeCursor( self, maxCount=4 ):
    """
    Works backwards from the cursor finding word characters
        (which we might then want to autocomplete).

    Needed for auto-complete functions.
    """
    #if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
        #vPrint( 'Quiet', debuggingThisModule, "AutocompleteFunctions.getWordCharactersBeforeCursor( {} )".format( maxCount ) )

    previousText = self.textBox.get( tk.INSERT+'-{}c'.format( maxCount ), tk.INSERT )
    #vPrint( 'Quiet', debuggingThisModule, "previousText", repr(previousText) )
    wordText = ''
    for previousChar in reversed( previousText ):
        if previousChar in self.autocompleteWordChars:
            wordText = previousChar + wordText
        else: break
    #vPrint( 'Quiet', debuggingThisModule, 'getWordCharactersBeforeCursor: returning wordText', repr(wordText) )
    return wordText
# end of AutocompleteFunctions.getWordCharactersBeforeCursor


def getCharactersAndWordBeforeCursor( self, maxCount=4 ):
    """
    Works backwards from the cursor finding word characters
        INCLUDING THE PREVIOUS WORD
        (which we might then want to autocomplete).

    Needed for auto-complete functions.
    """
    #if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
        #vPrint( 'Quiet', debuggingThisModule, "AutocompleteFunctions.getCharactersAndWordBeforeCursor( {} )".format( maxCount ) )

    previousText = self.textBox.get( tk.INSERT+'-{}c'.format( maxCount ), tk.INSERT )
    #vPrint( 'Quiet', debuggingThisModule, "previousText", repr(previousText) )
    delimiterCount = 0
    wordText = ''
    for previousChar in reversed( previousText ):
        if previousChar in self.autocompleteWordChars:
            wordText = previousChar + wordText
        elif previousChar in BibleOrgSysGlobals.TRAILING_WORD_END_CHARS+MULTIPLE_SPACE_SUBSTITUTE+TRAILING_SPACE_SUBSTITUTE:
            if delimiterCount > 0: break
            #vPrint( 'Quiet', debuggingThisModule, "Found delimiter {!r}".format( previousChar ) )
            wordText = previousChar + wordText
            delimiterCount += 1
    #vPrint( 'Quiet', debuggingThisModule, 'getCharactersAndWordBeforeCursor: returning wordText', repr(wordText) )
    return wordText
# end of AutocompleteFunctions.getCharactersAndWordBeforeCursor


def getWordBeforeSpace( self, maxCount=20 ):
    """
    Works backwards from before the word ending character (e.g., a space) before the cursor
        trying to find the word that was last entered.

    Needed for auto-complete functions.
    """
    #if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
        #vPrint( 'Quiet', debuggingThisModule, "AutocompleteFunctions.getWordBeforeSpace( {} )".format( maxCount ) )

    previousText = self.textBox.get( tk.INSERT+'-{}c'.format( maxCount ), tk.INSERT )
    #vPrint( 'Quiet', debuggingThisModule, "previousText1", repr(previousText) )
    assert previousText and previousText[-1] in BibleOrgSysGlobals.TRAILING_WORD_END_CHARS+MULTIPLE_SPACE_SUBSTITUTE+TRAILING_SPACE_SUBSTITUTE
    previousText = previousText[:-1] # Drop the character that ended the word
    #vPrint( 'Quiet', debuggingThisModule, "previousText2", repr(previousText) )
    wordText = ''
    if 1 or previousText and previousText[-1].isalpha():
        for previousChar in reversed( previousText ):
            if previousChar in self.autocompleteWordChars:
                wordText = previousChar + wordText
            else: break
    #vPrint( 'Quiet', debuggingThisModule, 'getWordBeforeSpace: returning word Text', repr(wordText) )
    return wordText
# end of AutocompleteFunctions.getWordBeforeSpace


def acceptAutocompleteSelection( self, includeTrailingSpace=False ):
    """
    Used by autocomplete routines in onTextChange.

    Gets the chosen word and inserts the end of it into the text.
    """
    if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
        #vPrint( 'Quiet', debuggingThisModule, "AutocompleteFunctions.acceptAutocompleteSelection( {} )".format( includeTrailingSpace ) )
        assert self.autocompleteBox is not None

    currentWord = self.autocompleteBox.get( tk.ACTIVE )
    #vPrint( 'Quiet', debuggingThisModule, '  autocompleteBox currentWord', currentWord )
    self.removeAutocompleteBox()

    if self.autocompleteOverlap:
        #vPrint( 'Quiet', debuggingThisModule, "Have {!r} with overlap {!r}".format( currentWord, self.autocompleteOverlap ) )
        assert currentWord.startswith( self.autocompleteOverlap )
        #currentWord = currentWord[len(self.autocompleteOverlap):]

    # Autocomplete by inserting the rest of the selected word plus a space
    # NOTE: The user has to backspace over the space if they don't want it (e.g., to put a period)
    # NOTE: The box reappears with the current code if we don't append the space -- would need to add a flag
    self.textBox.insert( tk.INSERT, currentWord[len(self.autocompleteOverlap):] \
                                    + (' ' if includeTrailingSpace else '') )

    #vPrint( 'Quiet', debuggingThisModule, "acceptAutocompleteSelection for {!r}".format( currentWord ) )
    addNewAutocompleteWord( self, currentWord )

    ## Put this word at the beginning of the list so it comes up on top next time
    #firstLetter, remainder = currentWord[0], currentWord[1:]
    #self.autocompleteWords[firstLetter].remove( remainder )
    #self.autocompleteWords[firstLetter].insert( 0, remainder )
# end of AutocompleteFunctions.acceptAutocompleteSelection


def addNewAutocompleteWord( self, possibleNewWord ):
    """
    Add the new autocomplete word if necessary,
        or at least bring it to the top of the list.

    Used by autocomplete routines in onTextChange.
    """
    if BibleOrgSysGlobals.debugFlag and debuggingThisModule:
        vPrint( 'Quiet', debuggingThisModule, "AutocompleteFunctions.addNewAutocompleteWord( {!r} )".format( possibleNewWord ) )
        assert isinstance( possibleNewWord, str )
        assert possibleNewWord

    if ' ' in possibleNewWord: # bring each separate word to the top
        remainder = possibleNewWord
        while ' ' in remainder:
            individualWord, remainder = remainder.split( None, 1 )
            #vPrint( 'Quiet', debuggingThisModule, "  word={!r}, remainder={!r}".format( individualWord, remainder ) )
            #vPrint( 'Quiet', debuggingThisModule, "Recursive1 of {!r}".format( individualWord ) )
            addNewAutocompleteWord( self, individualWord ) # recursive call
            #vPrint( 'Quiet', debuggingThisModule, "Recursive2 of {!r}".format( remainder ) )
            addNewAutocompleteWord( self, remainder ) # recursive call

    while possibleNewWord and possibleNewWord[-1] in END_CHARS_TO_REMOVE:
        possibleNewWord = possibleNewWord[:-1] # Remove certain final punctuation

    if len( possibleNewWord ) > self.autocompleteMinLength:
        #vPrint( 'Quiet', debuggingThisModule, "Adding new autocomplete word: {!r}".format( possibleNewWord ) )
        # Put this word at the beginning of the list so it comes up on top next time
        firstLetter, remainder = possibleNewWord[0], possibleNewWord[1:]
        try: self.autocompleteWords[firstLetter].remove( remainder )
        except ValueError: pass # remove will fail if this really is a new word
        except KeyError: # There's no list existing for this letter
            self.autocompleteWords[firstLetter] = []
        self.autocompleteWords[firstLetter].insert( 0, remainder )
# end of AutocompleteFunctions.addNewAutocompleteWord




def briefDemo() -> None:
    """
    Demo program to handle command line parameters and then run what they want.
    """
    BibleOrgSysGlobals.introduceProgram( __name__, programNameVersion, LAST_MODIFIED_DATE )
    if BibleOrgSysGlobals.debugFlag: vPrint( 'Quiet', debuggingThisModule, "Running demo…" )

    tkRootWindow = tk.Tk()
    tkRootWindow.title( programNameVersion )
    tkRootWindow.textBox = tk.Text( tkRootWindow )

    #uEW = AutocompleteFunctions( tkRootWindow, None )

    # Start the program running
    # tkRootWindow.mainloop()
# end of AutocompleteFunctions.briefDemo

def fullDemo() -> None:
    """
    Full demo to check class is working
    """
    BibleOrgSysGlobals.introduceProgram( __name__, programNameVersion, LAST_MODIFIED_DATE )
    if BibleOrgSysGlobals.debugFlag: vPrint( 'Quiet', debuggingThisModule, "Running demo…" )

    tkRootWindow = tk.Tk()
    tkRootWindow.title( programNameVersion )
    tkRootWindow.textBox = tk.Text( tkRootWindow )

    #uEW = AutocompleteFunctions( tkRootWindow, None )

    # Start the program running
    tkRootWindow.mainloop()
# end of AutocompleteFunctions.fullDemo

if __name__ == '__main__':
    from multiprocessing import freeze_support
    freeze_support() # Multiprocessing support for frozen Windows executables

    # Configure basic set-up
    parser = BibleOrgSysGlobals.setup( SHORT_PROGRAM_NAME, PROGRAM_VERSION, LAST_MODIFIED_DATE )
    BibleOrgSysGlobals.addStandardOptionsAndProcess( parser )

    fullDemo()

    BibleOrgSysGlobals.closedown( PROGRAM_NAME, PROGRAM_VERSION )
# end of AutocompleteFunctions.py
